package org.processmining.plugins.actitrac;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.ScrollPaneConstants;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.actitrac.ui.JClusteringInfo;
import org.processmining.plugins.actitrac.ui.JClusteringPane;
import org.processmining.plugins.actitrac.ui.JClusteringSidePane;


/**
 * Plugin for the ProM framework to construct a graphical representation of an
 * ActiTraCClustering object.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-07-29
 */
@Plugin(
		name = "ActiTraCClusteringVisualiser",
		returnLabels = "Clustering",
		returnTypes = JComponent.class,
		parameterLabels = { "ActiTraCClustering" },
		userAccessible = true)
@Visualizer
public class ActiTraCClusteringVisualiser {

	@PluginVariant(
			variantLabel = "ActiTraCClusteringVisualiser",
			requiredParameterLabels = { 0 })
	public JComponent visualise(UIPluginContext context,
			ActiTraCClustering clustering) {
		/*
		 * Visual hierarchy:
		 * 
		 * visualisation
		 * + scrollpaneParent [center]
		 * | + scrollpane [back layer; entire panel]
		 * | + clusteringinfo [back layer; bottom]
		 * | + clusterinfo [top layer; right side]
		 * + horizontal scrollbar [bottom]
		 * + vertical scrollbar [right]
		 */

		// Initialise some components
		JPanel visualisation = new JPanel();
		JLayeredPane scrollpaneParent = new JLayeredPane();
		final JScrollPane scrollpane = new JScrollPane();
		JViewport viewport = new JViewport();
		final JClusteringInfo clusteringinfo = new JClusteringInfo(clustering);
		final JClusteringSidePane clusterinfo = new JClusteringSidePane(context,
				clustering);
		JClusteringPane clusteringpane = new JClusteringPane(clustering,
				clusterinfo);

		visualisation.setLayout(new BorderLayout());

		// Make sure the backgrounds of these components are not drawn. This
		// will ensure that they will not interfere with visualisation's
		// background colour.
		scrollpane.setOpaque(false);
		viewport.setOpaque(false);
		scrollpaneParent.setOpaque(false);

		visualisation.setBackground(Color.WHITE);
		scrollpane
				.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollpane
				.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER);
		scrollpane.getVerticalScrollBar().setUnitIncrement(16);
		scrollpane.setBorder(BorderFactory.createEmptyBorder());

		// Create the hierarchy of components
		viewport.add(clusteringpane);
		scrollpane.setViewport(viewport);
		scrollpaneParent.add(scrollpane, JLayeredPane.DEFAULT_LAYER);
		scrollpaneParent.add(clusteringinfo, JLayeredPane.DEFAULT_LAYER);
		scrollpaneParent.add(clusterinfo, JLayeredPane.PALETTE_LAYER);
		visualisation.add(scrollpaneParent, BorderLayout.CENTER);
		visualisation.add(scrollpane.getVerticalScrollBar(),
				BorderLayout.LINE_END);
		visualisation.add(scrollpane.getHorizontalScrollBar(),
				BorderLayout.PAGE_END);

		// Set the layout manager of scrollpaneParent so that
		// 1. scrollpane (on the lowest layer) takes up all available space
		// 2. clusteringinfo (on the lowest layer) takes up the bottom portion
		// of the panel (the height is determined by clusteringinfo's preferred
		// size)
		// 3. clusterinfo (on the top layer) takes up the right part of the
		// panel (the width is determined by clusterinfo's preferred size)
		scrollpaneParent.setLayout(new LayoutManager() {

			@Override
			public void addLayoutComponent(String arg0, Component arg1) {
			}

			@Override
			public void layoutContainer(Container parent) {
				scrollpane.setBounds(0, 0, parent.getWidth(),
						parent.getHeight());
				Dimension d2 = clusteringinfo.getPreferredSize();
				clusteringinfo.setBounds(0, parent.getHeight() - d2.height,
						d2.width, d2.height);
				Dimension d3 = clusterinfo.getPreferredSize();
				clusterinfo.setBounds(parent.getWidth() - d3.width, 0,
						d3.width, parent.getHeight());
			}

			@Override
			public Dimension minimumLayoutSize(Container arg0) {
				return scrollpane.getMinimumSize();
			}

			@Override
			public Dimension preferredLayoutSize(Container arg0) {
				return scrollpane.getPreferredSize();
			}

			@Override
			public void removeLayoutComponent(Component arg0) {
			}
		});

		// Connect the clusterinfo panel to the context so that it may use the
		// context after this method returns.
		// This is required to export any information from the visualisation
		// back to the framework.
		context.addConnection(JClusteringSidePaneConnection.plugin(clusterinfo));
		return visualisation;
	}
}
