package org.processmining.plugins.actitrac.algorithm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.types.GroupedXLog;

import cern.colt.matrix.DoubleMatrix2D;

/**
 * This class is an extension of the regular {@link Cluster}-class. It is able
 * to calculate a number of statistics based on a cluster and cache those.
 * Nearly every method in this class will only fully execute when it is called
 * for the first time. If it is called again afterwards, it will return the same
 * result as on the previous execution without requiring additional computation
 * time. Many methods in this class are synchronised and should be safe to be
 * used in multi-threaded applications.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-07-27
 */
public class ClusterStatistics {

	/**
	 * The cluster for which all statistics are to be calculated.
	 */
	private Cluster cluster;

	// The names of these fields ought to be clear enough.

	private int nbTraces;

	private boolean nbTracesCached = false;

	private int dpi;
	private boolean dpiCached = false;

	private int nbEvents;

	private boolean nbEventsCached = false;
	private int nbInternalDistinctEvents;

	private List<XEvent> listInternalDistinctEvents;

	private boolean nbInternalDistinctEventsCached = false;
	private int nbExternalDistinctEvents;

	private List<XEvent> listExternalDistinctEvents;

	private boolean nbExternalDistinctEventsCached = false;
	private int nbDistinctEvents;
	private List<XEvent> listDistinctEvents;

	private boolean nbDistinctEventsCached = false;

	private int nbExtraConformantTraces;

	private int nbExtraNonConformantTraces;
	private boolean nbExtraConformantTracesCached = false;
	private float pm;

	private int nbFittingDpi;

	private boolean pmCached = false;

	private int nbArcs;
	private boolean nbArcsCached = false;

	/**
	 * Create a ClusterStatistics object for the given Cluster.
	 */
	public ClusterStatistics(Cluster cluster) {
		this.cluster = cluster;
	}

	/**
	 * Return the Average Connector Degree. This is the average amount of arcs
	 * that enter or leave a node in the HeuristicsNet.
	 */
	public float getAverageConnectorDegree() {
		return getNbArcs() * 2 / (float) getNbNodes();
	}

	/**
	 * Return the Cluster on which we are operating.
	 */
	public Cluster getCluster() {
		return cluster;
	}

	/**
	 * Return the distinct events that are used in this cluster. Each distinct
	 * event occurs only once in this list. Elements are considered to be
	 * distinct when either their concept:name or lifecycle:transition differ.
	 * 
	 * @return A list of distinct events.
	 */
	public List<XEvent> getDistinctEvents() {
		if (!nbDistinctEventsCached)
			getNbDistinctEvents();
		return listDistinctEvents;
	}

	/**
	 * Return the distinct events that are part of the "external" traces of this
	 * cluster. Traces are "external" when they are part of a cluster, but not
	 * used to calculate the underlying process model for that cluster. Elements
	 * are considered to be distinct when either their concept:name or
	 * lifecycle:transition differ.
	 * 
	 * @return A list of distinct events.
	 */
	synchronized public List<XEvent> getExternalDistinctEvents() {
		if (!nbExternalDistinctEventsCached)
			getNbExternalDistinctEvents();
		return listExternalDistinctEvents;
	}

	/**
	 * Return the distinct events that are part of the "internal" traces of this
	 * cluster. Traces are "internal" when they are part of a cluster and used
	 * to create the underlying process model. Elements are considered to be
	 * distinct when either their concept:name or lifecycle:transition differ.
	 * 
	 * @return A list of distinct events.
	 */
	synchronized public List<XEvent> getInternalDistinctEvents() {
		if (!nbInternalDistinctEventsCached)
			getNbInternalDistinctEvents();
		return listInternalDistinctEvents;
	}

	/**
	 * Return the average amount of events of which a trace consists in this
	 * cluster.
	 */
	synchronized public float getMeanTraceLength() {
		return getNbEvents() / (float) getNbTraces();
	}

	/**
	 * Return the amount of arcs in the HeuristicsNet.
	 */
	synchronized public int getNbArcs() {
		if (nbArcsCached)
			return nbArcs;
		DoubleMatrix2D matrix = cluster.getProcessModel().getHeuristicsNet()
				.getArcUsage();
		nbArcs = matrix.cardinality();
		return nbArcs;
	}

	/**
	 * Return the amount of distinct events that are used in this cluster. Each
	 * distinct event occurs only once in this list. Elements are considered to
	 * be distinct when either their concept:name or lifecycle:transition
	 * differ.
	 */
	synchronized public int getNbDistinctEvents() {
		if (nbDistinctEventsCached)
			return nbDistinctEvents;
		HashMap<String, XEvent> events = new HashMap<String, XEvent>();
		for (XEvent event : getInternalDistinctEvents()) {
			XAttributeMap attributes = event.getAttributes();
			String identifier = attributes.get("concept:name").toString();
			identifier += '\0';
			identifier += attributes.get("lifecycle:transition").toString();
			if (!events.containsKey(identifier)) {
				events.put(identifier, event);
			}
		}
		for (XEvent event : getExternalDistinctEvents()) {
			XAttributeMap attributes = event.getAttributes();
			String identifier = attributes.get("concept:name").toString();
			identifier += '\0';
			identifier += attributes.get("lifecycle:transition").toString();
			if (!events.containsKey(identifier)) {
				events.put(identifier, event);
			}
		}
		listDistinctEvents = new ArrayList<XEvent>();
		for (Entry<String, XEvent> entry : events.entrySet()) {
			listDistinctEvents.add(entry.getValue());
		}
		nbDistinctEvents = listDistinctEvents.size();
		nbDistinctEventsCached = true;
		return nbDistinctEvents;
	}

	/**
	 * Return the amount of distinct process instances in this cluster. Process
	 * instances are equivalent to traces. Two process instances are considered
	 * to be distinct when the i-th event in the first process instance is not
	 * the same as the i-th event in the second process instance. Events are
	 * defined as being the same when their concept:name and
	 * lifecycle:transition both are the same.
	 */
	synchronized public int getNbDpi() {
		if (dpiCached)
			return dpi;
		XLog xlog = cluster.getInternalLog();
		xlog.addAll(cluster.getExternalLog());
		GroupedXLog glog = new GroupedXLog(xlog);
		dpi = glog.size();
		dpiCached = true;
		return dpi;
	}

	/**
	 * Return the amount of events in this cluster. This is the sum of the
	 * amount of events of each trace.
	 */
	synchronized public int getNbEvents() {
		if (nbEventsCached)
			return nbEvents;
		int nbEvents = 0;
		for (XTrace trace : cluster.getInternalLog())
			nbEvents += trace.size();
		for (XTrace trace : cluster.getExternalLog())
			nbEvents += trace.size();
		this.nbEvents = nbEvents;
		nbEventsCached = true;
		return nbEvents;
	}

	/**
	 * Return the amount of distinct events that are part of the "external"
	 * traces of this cluster. Traces are "external" when they are part of a
	 * cluster, but not used to calculate the underlying process model for that
	 * cluster. Elements are considered to be distinct when either their
	 * concept:name or lifecycle:transition differ.
	 */
	synchronized public int getNbExternalDistinctEvents() {
		if (nbExternalDistinctEventsCached)
			return nbExternalDistinctEvents;
		HashMap<String, XEvent> events = new HashMap<String, XEvent>();
		for (XTrace trace : cluster.getExternalLog()) {
			for (XEvent event : trace) {
				XAttributeMap attributes = event.getAttributes();
				String identifier = attributes.get("concept:name").toString();
				identifier += '\0';
				identifier += attributes.get("lifecycle:transition").toString();
				if (!events.containsKey(identifier)) {
					events.put(identifier, event);
				}
			}
		}
		listExternalDistinctEvents = new ArrayList<XEvent>();
		for (Entry<String, XEvent> entry : events.entrySet()) {
			listExternalDistinctEvents.add(entry.getValue());
		}
		nbExternalDistinctEvents = listExternalDistinctEvents.size();
		nbExternalDistinctEventsCached = true;
		return nbExternalDistinctEvents;
	}

	/**
	 * Return the amount of traces that
	 * <ol>
	 * <li>are in this cluster</li>
	 * <li>are not used to build the underlying process model</li>
	 * <li>perfectly fit the underlying process model</li>
	 * </ol>
	 */
	synchronized public int getNbExtraConformantTraces() {
		if (nbExtraConformantTracesCached)
			return nbExtraConformantTraces;
		ProcessModel model = cluster.getProcessModel();
		int nbExtraConformantTraces = 0;
		int nbExtraNonConformantTraces = 0;
		for (XTrace trace : cluster.getExternalLog()) {
			if (model.traceFitness(trace) == 1.0)
				nbExtraConformantTraces++;
			else
				nbExtraNonConformantTraces++;
		}
		this.nbExtraConformantTraces = nbExtraConformantTraces;
		this.nbExtraNonConformantTraces = nbExtraNonConformantTraces;
		nbExtraConformantTracesCached = true;
		return nbExtraConformantTraces;
	}

	/**
	 * Return the amount of traces that
	 * <ol>
	 * <li>are in this cluster</li>
	 * <li>are not used to build the underlying process model</li>
	 * <li>don't (perfectly) fit the underlying process model</li>
	 * </ol>
	 */
	synchronized public int getNbExtraNonConformantTraces() {
		if (!nbExtraConformantTracesCached)
			getNbExtraConformantTraces();
		return nbExtraNonConformantTraces;
	}

	/**
	 * Return the amount of distinct process instances in this cluster and fit
	 * the underlying process model. Process instances are equivalent to traces.
	 * Two process instances are considered to be distinct when the i-th event
	 * in the first process instance is not the same as the i-th event in the
	 * second process instance. Events are defined as being the same when their
	 * concept:name and lifecycle:transition both are the same.
	 */
	synchronized public int getNbFittingDpi() {
		if (!pmCached)
			getPM();
		return nbFittingDpi;
	}

	/**
	 * Return the amount of distinct events that are part of the "internal"
	 * traces of this cluster. Traces are "internal" when they are part of a
	 * cluster, and used to calculate the underlying process model for that
	 * cluster. Elements are considered to be distinct when either their
	 * concept:name or lifecycle:transition differ.
	 */
	synchronized public int getNbInternalDistinctEvents() {
		if (nbInternalDistinctEventsCached)
			return nbInternalDistinctEvents;
		HashMap<String, XEvent> events = new HashMap<String, XEvent>();
		for (XTrace trace : cluster.getInternalLog()) {
			for (XEvent event : trace) {
				XAttributeMap attributes = event.getAttributes();
				String identifier = attributes.get("concept:name").toString();
				identifier += '\0';
				identifier += attributes.get("lifecycle:transition").toString();
				if (!events.containsKey(identifier)) {
					events.put(identifier, event);
				}
			}
		}
		listInternalDistinctEvents = new ArrayList<XEvent>();
		for (Entry<String, XEvent> entry : events.entrySet()) {
			listInternalDistinctEvents.add(entry.getValue());
		}
		nbInternalDistinctEvents = listInternalDistinctEvents.size();
		nbInternalDistinctEventsCached = true;
		return nbInternalDistinctEvents;
	}

	/**
	 * Return the amount of nodes of which the HeuristicsNet associated with
	 * this cluster consists.
	 */
	public int getNbNodes() {
		return cluster.getProcessModel().getHeuristicsNet().size();
	}

	/**
	 * Return the total amount of traces in this cluster.
	 */
	synchronized public int getNbTraces() {
		if (nbTracesCached)
			return nbTraces;
		int nbTraces = cluster.getInternalLog().size();
		nbTraces += cluster.getExternalLog().size();
		this.nbTraces = nbTraces;
		nbTracesCached = true;
		return nbTraces;
	}

	/**
	 * Return the density of the HeuristicsNet associated with this cluster. The
	 * density is defined as the amount of arcs over the maximal possible amount
	 * of arcs.
	 * 
	 * The maximal possible amount of arcs is the amount of nodes squared. (Each
	 * node can be connected to every other node and itself.)
	 */
	public float getNetDensity() {
		float arcs = getNbArcs();
		float nodes = getNbNodes();
		return arcs / (nodes * nodes);
	}

	/**
	 * Return the Parsing Measure of this cluster. The parsing measure is
	 * defined as the amount of traces that fit the underlying process model
	 * over the total amount of traces in the cluster.
	 */
	synchronized public float getPM() {
		if (pmCached)
			return pm;
		int nbFittingDpi = 0;
		float pm = 0;
		ProcessModel model = cluster.getProcessModel();
		XLog xlog = cluster.getInternalLog();
		xlog.addAll(cluster.getExternalLog());
		GroupedXLog glog = new GroupedXLog(xlog);
		for (int i = 0; i < glog.size(); i++) {
			XTrace trace = glog.get(i).get(0);
			if (model.traceFitness(trace) == 1.0) {
				nbFittingDpi++;
				pm += glog.get(i).size();
			}
		}
		pm /= getNbTraces();
		this.pm = pm;
		this.nbFittingDpi = nbFittingDpi;
		this.pmCached = true;
		return pm;
	}
}
