package org.processmining.plugins.actitrac.algorithm;

import org.processmining.plugins.actitrac.ActiTraCClustering;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.utils.ATCParameters;
import org.processmining.plugins.actitrac.utils.ClustererEventListener;
import org.processmining.plugins.actitrac.utils.OperationCancelledException;
import org.processmining.plugins.heuristicsnet.miner.heuristics.miner.settings.HeuristicsMinerSettings;


/**
 * Abstract type representing classes that will be able to cluster an log into
 * an ActiTraCClustering. This class already implements everything required for
 * the use of a ClustererEventListener. It also keeps track of ATCParameters,
 * HeuristicsMinerSettings and the generated ActiTraCClustering-object.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Joch-en De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-03
 */
public abstract class Clusterer {

	/**
	 * The resulting clustering. This object is initiated by {@link #cluster()}.
	 */
	private ActiTraCClustering clustering;

	/**
	 * The ActiTraC-parameters we have to follow.
	 */
	protected ATCParameters parameters;

	/**
	 * The HeuristicsMinerSettings we should give to HeuristicsMiner.
	 */
	protected HeuristicsMinerSettings settings;

	/**
	 * A ClustererEventListener.
	 */
	private ClustererEventListener eventListener = null;

	/**
	 * Create a new Clusterer to process the given GroupedXLog, while taking the
	 * given parameters and settings in account.
	 * 
	 * @param log
	 *            The log to be clustered.
	 * @param parameters
	 *            The ActiTraC-parameters we are to follow.
	 * @param settings
	 *            The settings we should give to HeuristicsMiner.
	 */
	public Clusterer(ATCParameters parameters, HeuristicsMinerSettings settings) {
		this(parameters, settings, null);
	}

	/**
	 * Create a new Clusterer to process the given GroupedXLog, while taking the
	 * given parameters and settings in account.
	 * 
	 * @param log
	 *            The log to be clustered.
	 * @param parameters
	 *            The ActiTraC-parameters we are to follow.
	 * @param settings
	 *            The settings we should give to HeuristicsMiner.
	 * @param l
	 *            The ClustererEventListener that should be assigned to this
	 *            Clusterer.
	 */
	public Clusterer(ATCParameters parameters,
			HeuristicsMinerSettings settings, ClustererEventListener l) {
		this.parameters = parameters;
		this.settings = settings;
		this.eventListener = l;
	}

	/**
	 * Adds the given ClustererEventListener to this Clusterer. If an
	 * clusterEventListener was already assigned to this Clusterer, it will be
	 * removed. Pass null as argument will remove the current listener.
	 * 
	 * @param l
	 *            The new listener, or null.
	 */
	public void addClusterEventListener(ClustererEventListener l) {
		eventListener = l;
	}

	/**
	 * Calls {@link ClustererEventListener#cancel()} on the event listener and
	 * then throws an OperationCancelledException.
	 */
	protected void cancel() throws OperationCancelledException {
		if (eventListener != null)
			eventListener.cancel();
		throw new OperationCancelledException();
	}

	/**
	 * Perform the actual processing on the log.
	 * 
	 * @throws OperationCancelledException
	 *             When the ClusterEventListener that has been assigned to this
	 *             Clusterer (if any) suggests to cancel this operation.
	 */
	public abstract void cluster() throws OperationCancelledException;

	/**
	 * Calls {@link ClustererEventListener#clusterCreated(c)} on the event
	 * listener.
	 * 
	 * @param c
	 *            The newly created cluster.
	 */
	protected void fireClusterCreatedEvent(Cluster c) {
		if (eventListener != null)
			eventListener.clusterCreated(c);
	}

	/**
	 * Calls {@link ClustererEventListener#end()} on the event listener.
	 */
	protected void fireEndEvent() {
		if (eventListener != null)
			eventListener.end();
	}

	/**
	 * Calls {@link ClustererEventListener#minimalClusterSizeImpossible()} on
	 * the event listener.
	 * 
	 * @return True if the Clusterer should continue, false otherwise.
	 */
	protected boolean fireMinimalClusterSizeImpossible() {
		if (eventListener != null)
			return eventListener.minimalClusterSizeImpossible();
		return true;
	}

	/**
	 * Calls {@link ClustererEventListener#nonfittingAssigned()} on the event
	 * listener.
	 */
	protected void fireNonfittingAssigned() {
		if (eventListener != null)
			eventListener.nonfittingAssigned();
	}

	/**
	 * Calls {@link ClustererEventListener#start()} on the event listener.
	 */
	protected void fireStartEvent() {
		if (eventListener != null)
			eventListener.start();
	}

	/**
	 * Calls {@link ClustererEventListener#targetReached(left)} on the event
	 * listener.
	 * 
	 * @param left
	 *            The amount of similar XTrace groups that have not yet been
	 *            assigned to a cluster.
	 */
	protected void fireTargetReachedEvent(int left) {
		if (eventListener != null)
			eventListener.targetReached(left);
	}

	/**
	 * Return the ActiTraCClustering resulting from the calculations done by
	 * this object. The result of this method is not defined before the method
	 * {@link #cluster()} has been called on this object.
	 */
	public ActiTraCClustering getClustering() {
		return clustering;
	}

	/**
	 * Set the resulting ActiTraCClustering of this Clusterer to the given
	 * ActiTraCClustering.
	 */
	protected void setClustering(ActiTraCClustering clustering) {
		this.clustering = clustering;
	}

	/**
	 * Checks whether this Clusterer should cancel. This is done through the
	 * method {@link ClustererEventListener#shouldCancel()}. This method should
	 * return true if this clusterer should cancel.
	 */
	protected boolean shouldCancel() {
		if (eventListener != null)
			return eventListener.shouldCancel();
		else
			return false;
	}

	/**
	 * Calls {@link ClustererEventListener#importing()} on the event listener.
	 */
	protected void fireImportingEvent() {
		if (eventListener != null)
			eventListener.importing();
	}
}
