package org.processmining.plugins.actitrac.algorithm;

import java.util.ArrayList;
import java.util.List;

import org.deckfour.xes.model.XLog;
import org.processmining.plugins.actitrac.ActiTraCClustering;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.types.ClusterTree;
import org.processmining.plugins.actitrac.types.GroupedXLog;
import org.processmining.plugins.actitrac.utils.ATCParameters;
import org.processmining.plugins.actitrac.utils.ClustererEventListener;
import org.processmining.plugins.actitrac.utils.OperationCancelledException;
import org.processmining.plugins.heuristicsnet.miner.heuristics.miner.settings.HeuristicsMinerSettings;


/**
 * This class is able to cluster an XLog. It will use the ability of
 * HeuristicsMiner to create a correct process model to cluster the log. The
 * clustering will happen from most common distinct process instance to least
 * common distinct process instance.
 * 
 * The clustering will begin with the most frequent dpi. Then the remaining dpi
 * are considered from most frequent to least frequent. When possible, a dpi is
 * added to the model. When this is impossible, that dpi will be skipped, or the
 * clustering stops extending the process model. This decision depends on the
 * minimal cluster size.
 * 
 * Next, every dpi that is not yet clustered but fits the current model
 * perfectly will be added to the model as well. This process is repeated while
 * the stop criteria have not been reached. Then finally, the remaining dpi will
 * be added to a new cluster or to an existing one.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-04
 */
public class GreedyClusterer extends Clusterer {

	/**
	 * The log to be processed. This object should *not* be altered.
	 */
	private GroupedXLog log;

	/**
	 * The number of the cluster that is currently being calculated. This is
	 * used to create a name for the new cluster.
	 */
	protected int cluster_nb = 1;

	/**
	 * Create a new GreedyClusterer to process the given XLog, while taking the
	 * given parameters and settings in account.
	 * 
	 * @param log
	 *            The log to be clustered.
	 * @param parameters
	 *            The ActiTraC-parameters we are to follow.
	 * @param settings
	 *            The settings we should give to HeuristicsMiner.
	 */
	public GreedyClusterer(XLog log, ATCParameters parameters,
			HeuristicsMinerSettings settings) {
		this(log, parameters, settings, null);
	}

	/**
	 * Create a new GreedyClusterer to process the given XLog, while taking the
	 * given parameters and settings in account.
	 * 
	 * @param log
	 *            The log to be clustered.
	 * @param parameters
	 *            The ActiTraC-parameters we are to follow.
	 * @param settings
	 *            The settings we should give to HeuristicsMiner.
	 * @param l
	 *            The ClustererEventListener that should be assigned to this
	 *            Clusterer.
	 */
	public GreedyClusterer(XLog log, ATCParameters parameters,
			HeuristicsMinerSettings settings, ClustererEventListener l) {
		super(parameters, settings, l);
		fireImportingEvent();
		this.log = new GroupedXLog(log);
	}

	@Override
	public void cluster() throws OperationCancelledException {
		boolean stop_reached = false;
		GroupedXLog log = (GroupedXLog) this.log.clone();
		XLog original_log = this.log.getLog();
		Cluster pseudoCluster = new Cluster(original_log, new ProcessModel(
				original_log, settings), this.log.getName());
		ClusterTree clusterTree = new ClusterTree(pseudoCluster);
		int doneTraces = 0;
		fireStartEvent();
		int clustersCreated = 0;

		while (!stop_reached) {
			// PHASE ONE
			int traces_added = 0;
			int groups_count = 0;
			List<Integer> goodTraces = new ArrayList<Integer>();
			goodTraces.add(0);
			int current = 1;
			// Go over each set of traces. Try to build a model, while including
			// them. As soon as the created model has a fitness below the
			// threshold, break out of the loop without including the traces
			// that make the model fail.
			// Unless the minimal cluster size is not achieved, in which case
			// the remaining groups are to be processed until the cluster size
			// is guaranteed. If impossible, notify the listener, then continue.
			while (current < log.size()) {
				if (shouldCancel())
					cancel();
				goodTraces.add(current);
				XLog merged_traces = log.getLog(goodTraces);
				ProcessModel model = new ProcessModel(merged_traces, settings);
				if (model.getFitness() >= parameters.getTargetFitness()) {
					// Model is still good enough
					traces_added += log.get(current).size();
					current++;
					groups_count++;
				} else {
					// The model is no longer good enough
					goodTraces.remove(goodTraces.size() - 1);
					if (traces_added >= parameters.getMinimalClusterSize()
							* log.getNbTraces())
						// Stop if the minimal cluster size has been reached
						break;
					else {
						// Try again from the next group if the minimal cluster
						// size has not been reached
						current++;
						if (current >= log.size()) {
							if (!fireMinimalClusterSizeImpossible())
								throw new OperationCancelledException();
						}
					}
				}
			}

			XLog merged_traces = log.getLog(goodTraces);
			ProcessModel model = new ProcessModel(merged_traces, settings);
			// Create a Cluster, storing the ProcessModel and the XLog.
			Cluster newCluster = new Cluster(merged_traces, model,
					nextClusterName());
			// Remove all of the traces that have just been assigned to a
			// cluster from the list.
			for (int i = goodTraces.size() - 1; i >= 0; i--) {
				doneTraces += log.get(goodTraces.get(i)).size();
				log.remove(goodTraces.get(i));
			}
			current = 0;
			// PHASE ONE END

			// PHASE TWO
			// Go over each of the remaining traces. If they fit, add them to
			// the cluster as external traces.
			while (current < log.size()) {
				if (shouldCancel())
					cancel();
				if (model.traceFitness(log.get(current).get(0)) == 1.0f) {
					newCluster.addExternal(log.getLog(current));
					log.remove(current);
					current--;
				}
				current++;
			}
			clusterTree.addCluster(newCluster);
			fireClusterCreatedEvent(newCluster);
			// PHASE TWO END

			clustersCreated++;
			// While the stop criteria are not reached, restart at phase one
			if (parameters.getStopByNbGroups()) {
				int target = parameters.getStopNbGroups();
				if (parameters.getSeparateNonFitting())
					target--;
				if (target <= clustersCreated)
					stop_reached = true;
			} else if (parameters.getStopByPercentage()) {
				int totalTraces = this.log.getNbTraces();
				float done = (float) doneTraces / (float) totalTraces * 100f;
				if (done >= parameters.getStopPercentage())
					stop_reached = true;
			} else {
				stop_reached = true;
			}
			// If nothing is left, quit anyway
			if (log.size() == 0) {
				break;
			}
		}
		fireTargetReachedEvent(log.size());
		// Process the remaining traces
		if (parameters.getSeparateNonFitting()) {
			// Add the remaining traces to a new cluster
			XLog rubbish_log = null;
			for (int i = 0; i < log.size(); i++) {
				if (rubbish_log == null)
					rubbish_log = log.getLog(i);
				else
					rubbish_log.addAll(log.getLog(i));
			}
			if (rubbish_log != null) {
				ProcessModel rubbish_model = new ProcessModel(rubbish_log,
						settings);
				Cluster newCluster = new Cluster(rubbish_log, rubbish_model,
						this.log.getName() + "_other");
				fireClusterCreatedEvent(newCluster);
				clusterTree.addCluster(newCluster);
			}
		} else {
			// PHASE THREE
			// Add the remaining traces to existing clusters
			for (int i = 0; i < log.size(); i++) {
				XLog group = log.getLog(i);
				float max_fitness = Float.NEGATIVE_INFINITY;
				Cluster best_fitness = null;
				for (Cluster current : clusterTree) {
					if (shouldCancel())
						cancel();
					float fitness = current.getProcessModel().traceFitness(
							group.get(0));
					if (fitness > max_fitness) {
						max_fitness = fitness;
						best_fitness = current;
					}
				}
				best_fitness.addExternal(group);
				fireNonfittingAssigned();
			}
			// PHASE THREE END
		}
		setClustering(new ActiTraCClustering(this.log, clusterTree, parameters,
				settings));
		fireEndEvent();
	}

	/**
	 * Calculate and return the name of the next cluster. This creates names of
	 * the form {original name}_{number}, where {number} is the amount of times
	 * this method has been called.
	 */
	private String nextClusterName() {
		StringBuilder sb = new StringBuilder();
		sb.append(log.getName());
		sb.append('_');
		if (cluster_nb < 10)
			sb.append('0');
		sb.append(cluster_nb);
		cluster_nb++;
		return sb.toString();
	}
}
