package org.processmining.plugins.actitrac.algorithm;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.deckfour.xes.model.XLog;
import org.processmining.plugins.actitrac.ActiTraCClustering;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.types.ClusterTree;
import org.processmining.plugins.actitrac.types.DistanceXLog;
import org.processmining.plugins.actitrac.types.GroupedXLog;
import org.processmining.plugins.actitrac.utils.ATCParameters;
import org.processmining.plugins.actitrac.utils.ClustererEventListener;
import org.processmining.plugins.actitrac.utils.OperationCancelledException;
import org.processmining.plugins.heuristicsnet.miner.heuristics.miner.settings.HeuristicsMinerSettings;

import org.processmining.plugins.actitrac.types.MaxRepeatAlphabetDistanceXLog;


/**
 * This class is able to cluster an XLog. It will use the ability of
 * HeuristicsMiner to create a correct process model to cluster the log. The
 * clustering of this Clusterer will start with the most frequent dpi. From
 * there on, within a certain sliding window determined by the greediness of the
 * algorithm, the dpi closest to the already existing set of dpi is considered.
 * When possible, this dpi is added. If this would create a bad process model,
 * either that dpi will be skipped, or the clustering stops extending the
 * process model. This decision depends on the minimal cluster size.
 * 
 * Next, every dpi that is not yet clustered but fits the current model
 * perfectly will be added to the model as well. This process is repeated while
 * the stop criteria have not been reached. Then finally, the remaining dpi will
 * be added to a new cluster or to an existing one.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-05
 */
public class GreedyDistanceClusterer extends Clusterer {

	/**
	 * The log to be processed. This object should *not* be altered.
	 */
	private GroupedXLog groupedLog;

	/**
	 * The log to be processed, with known distance between each two dpi.
	 */
	private DistanceXLog distanceLog;

	/**
	 * Current cluster being processed. Should only be altered by
	 * {@link #cluster()}.
	 */
	private int cluster_nb = 0;

	/**
	 * Create a new GreedyDistanceClusterer to process the given XLog, while
	 * taking the given parameters and settings in account.
	 * 
	 * @param log
	 *            The log to be clustered.
	 * @param parameters
	 *            The ActiTraC-parameters we are to follow.
	 * @param settings
	 *            The settings we should give to HeuristicsMiner.
	 * @throws IOException 
	 */
	public GreedyDistanceClusterer(XLog log, ATCParameters parameters,
			HeuristicsMinerSettings settings) throws IOException {
		this(log, parameters, settings, null);
	}

	/**
	 * Create a new GreedyDistanceClusterer to process the given XLog, while
	 * taking the given parameters and settings in account.
	 * 
	 * @param log
	 *            The log to be clustered.
	 * @param parameters
	 *            The ActiTraC-parameters we are to follow.
	 * @param settings
	 *            The settings we should give to HeuristicsMiner.
	 * @param l
	 *            The ClustererEventListener that should be assigned to this
	 *            Clusterer.
	 */
	public GreedyDistanceClusterer(XLog log, ATCParameters parameters,
			HeuristicsMinerSettings settings, ClustererEventListener l) throws IOException {
		super(parameters, settings, l);
		fireImportingEvent();
		groupedLog = new GroupedXLog(log);
		distanceLog = new MaxRepeatAlphabetDistanceXLog(groupedLog);;
	}

	@Override
	public void cluster() throws OperationCancelledException {
		// Take a copy of the grouped log, once a dpi is clustered, the
		// corresponding index in dpi will be set to null.
		XLog[] dpi = new XLog[groupedLog.size()];
		for (int i = 0; i < dpi.length; i++)
			dpi[i] = groupedLog.getLog(i);
		// Cluster containing the original status
		Cluster pseudoCluster = new Cluster(groupedLog.getLog(),
				new ProcessModel(groupedLog.getLog(), settings),
				groupedLog.getName());
		ClusterTree clusterTree = new ClusterTree(pseudoCluster);
		// Total amount of traces that have been processed
		int totalTracesDone = 0;
		fireStartEvent();

		// Stop already if you have nothing to work on
		boolean should_stop = dpi.length == 0;
		while (!should_stop) {
			// PHASE ONE
			cluster_nb++;
			// Take a copy of dpi, if a dpi has been rejected or added, the
			// index will be set to null.
			XLog[] dpi2 = dpi.clone();
			int nbTracesInDpi = 0;
			for (int i = 0; i < dpi.length; i++)
				if (dpi[i] != null)
					nbTracesInDpi += dpi[i].size();
			// List of dpi that are part of the cluster
			List<XLog> goodDpi = new ArrayList<XLog>();
			// List of the indices of the dpi that are part of the cluster
			List<Integer> goodDpiIndex = new ArrayList<Integer>();
			// Take the first dpi, this is always a part of the cluster
			for (int i = 0; i < dpi2.length; i++) {
				if (dpi2[i] == null)
					continue;
				goodDpi.add(dpi2[i]);
				goodDpiIndex.add(i);
				dpi2[i] = null;
				break;
			}
			// Create a new model
			ProcessModel model = new ProcessModel(goodDpi.get(0),
					super.settings);
			// Stop already if this one dpi doesn't give a good enough model
			boolean cluster_stop = model.getFitness() < parameters
					.getTargetFitness();
			// Amount of traces in the cluster
			int cluster_size = goodDpi.get(0).size();
			while (!cluster_stop) {
				if (shouldCancel())
					cancel();
				// Find next dpi to add
				int index = -1;
				{
					// Determine the minimal frequency a dpi must have
					int minimal_frequency = 0;
					int nb_dpi_left = 0;
					for (int j = 0; j < dpi2.length; j++)
						if (dpi2[j] != null)
							nb_dpi_left++;
					nb_dpi_left *= parameters.getDistanceGreediness();
					for (int j = 0; j < dpi2.length; j++) {
						if (dpi2[j] == null)
							continue;
						nb_dpi_left--;
						if (nb_dpi_left <= 0) {
							minimal_frequency = dpi2[j].size();
							break;
						}
					}
					// Find the closest dpi with a frequency of
					// minimal_frequency or above
					int minimal_distance = Integer.MAX_VALUE;
					int i = 0;
					while (i < dpi2.length) {
						// This dpi is already gone
						if (dpi2[i] == null) {
							i++;
							continue;
						}
						// We reached the infrequent dpi
						if (dpi2[i].size() < minimal_frequency)
							break;
						// Determine the distance from goodDpi to this dpi
						int distance = 0;
						for (Integer aGoodDpi : goodDpiIndex)
							distance += distanceLog.getDistance(aGoodDpi, i);
						// Update things if we found a better dpi
						if (distance < minimal_distance) {
							index = i;
							minimal_distance = distance;
						}
						i++;
					}
				}

				// No dpi found, we have to quit adding dpi's
				if (index == -1) {
					if (!fireMinimalClusterSizeImpossible())
						throw new OperationCancelledException();
					break;
				}

				// Try to add the found dpi
				XLog nextDpi = dpi2[index];
				ProcessModel newModel = model.add(nextDpi);
				if (newModel.getFitness() >= parameters.getTargetFitness()) {
					// The dpi fits well enough
					// Update cluster_size and model, add the dpi's log and
					// index, remove the dpi from dpi2
					cluster_size += nextDpi.size();
					goodDpi.add(nextDpi);
					goodDpiIndex.add(index);
					dpi2[index] = null;
					model = newModel;
				} else {
					// The dpi doesn't fit in the model
					// Minimal size a cluster should have
					int minimal_size = (int) (nbTracesInDpi * parameters
							.getMinimalClusterSize());
					if (cluster_size >= minimal_size) {
						// Minimal size reached, end the loop
						cluster_stop = true;
					} else {
						// Minimal size is not reached, remove from dpi2 and
						// loop again (ignore the dpi and continue)
						dpi2[index] = null;
					}
				}
			}
			// Remove all dpi in goodDpi from dpi
			for (int i = 0; i < goodDpiIndex.size(); i++)
				dpi[goodDpiIndex.get(i)] = null;
			// Create a Cluster, storing the ProcessModel and the XLog.
			XLog cluster_internal_log = (XLog) goodDpi.get(0).clone();
			for (int i = 1; i < goodDpi.size(); i++)
				cluster_internal_log.addAll(goodDpi.get(i));
			Cluster newCluster = new Cluster(cluster_internal_log, model,
					nextClusterName());
			// Update the count of clustered traces
			totalTracesDone += cluster_size;
			// END PHASE ONE

			// PHASE TWO
			// Go over each of the remaining traces. If they fit, add them to
			// the cluster as external traces.
			for (int i = 0; i < dpi.length; i++) {
				if (dpi[i] == null)
					continue;
				if (shouldCancel())
					cancel();
				if (model.traceFitness(dpi[i].get(0)) == 1.0f) {
					newCluster.addExternal(dpi[i]);
					totalTracesDone += dpi[i].size();
					dpi[i] = null;
				}
			}
			clusterTree.addCluster(newCluster);
			fireClusterCreatedEvent(newCluster);
			// PHASE TWO END

			// While the stop criteria are not reached, restart at phase one
			if (parameters.getStopByNbGroups()) {
				int target = parameters.getStopNbGroups();
				if (parameters.getSeparateNonFitting())
					target--;
				if (target <= cluster_nb) {
					should_stop = true;
				}
			} else if (parameters.getStopByPercentage()) {
				int totalTraces = this.groupedLog.getNbTraces();
				float done = (float) totalTracesDone / (float) totalTraces
						* 100f;
				if (done >= parameters.getStopPercentage())
					should_stop = true;
			} else {
				should_stop = true;
			}
			// If nothing is left, quit anyway
			boolean stuff_left = false;
			for (int i = 0; i < dpi.length; i++) {
				if (dpi[i] != null) {
					stuff_left = true;
					break;
				}
			}
			if (!stuff_left)
				should_stop = true;
		}
		int nbRemainingDpi = 0;
		for (int i = 0; i < dpi.length; i++)
			if (dpi[i] != null)
				nbRemainingDpi++;
		fireTargetReachedEvent(nbRemainingDpi);
		// Process the remaining traces
		if (parameters.getSeparateNonFitting()) {
			// Add the remaining traces to a new cluster
			XLog rubbish_log = null;
			for (int i = 0; i < dpi.length; i++) {
				if (dpi[i] == null)
					continue;
				if (rubbish_log == null) {
					rubbish_log = (XLog) dpi[i].clone();
					dpi[i] = null;
				} else {
					rubbish_log.addAll(dpi[i]);
					dpi[i] = null;
				}
			}
			if (rubbish_log != null) {
				ProcessModel rubbish_model = new ProcessModel(rubbish_log,
						settings);
				Cluster newCluster = new Cluster(rubbish_log, rubbish_model,
						this.groupedLog.getName() + "_other");
				fireClusterCreatedEvent(newCluster);
				clusterTree.addCluster(newCluster);
			}
		} else {
			// PHASE THREE
			// Add the remaining traces to existing clusters
			for (int i = 0; i < dpi.length; i++) {
				if (dpi[i] == null)
					continue;
				XLog group = dpi[i];
				float max_fitness = Float.NEGATIVE_INFINITY;
				Cluster best_fitness = null;
				for (Cluster current : clusterTree) {
					if (shouldCancel())
						cancel();
					float fitness = current.getProcessModel().traceFitness(
							group.get(0));
					if (fitness > max_fitness) {
						max_fitness = fitness;
						best_fitness = current;
					}
				}
				best_fitness.addExternal(group);
				fireNonfittingAssigned();
				dpi[i] = null;
			}
			// PHASE THREE END
		}
		setClustering(new ActiTraCClustering(this.groupedLog, clusterTree,
				parameters, settings));
		fireEndEvent();
	}

	/**
	 * Calculate and return the name of the next cluster. This creates names of
	 * the form {original name}_{number}, where {number} is the amount of times
	 * this method has been called.
	 */
	private String nextClusterName() {
		StringBuilder sb = new StringBuilder();
		sb.append(groupedLog.getName());
		sb.append('_');
		if (cluster_nb < 10)
			sb.append('0');
		sb.append(cluster_nb);
		return sb.toString();
	}
}
