package org.processmining.plugins.actitrac.algorithm;

import java.util.Random;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.models.heuristics.impl.ContinuousSemanticsParser;
import org.processmining.plugins.heuristicsnet.miner.heuristics.miner.settings.HeuristicsMinerSettings;

import org.processmining.plugins.heuristicsminer.HeuristicsMinerNoUI;

/**
 * Represents an XLog and its process model. The process model is a
 * HeuristicsNet, created by HeuristicsMiner.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-07-29
 */
public class ProcessModel {

	/**
	 * Calculate the fitness of the given trace in the given net, while updating
	 * the net.
	 */
	public static float traceFitness(XTrace trace, HeuristicsNet net) {
		ContinuousSemanticsParser parser = new ContinuousSemanticsParser(net,
				new Random(Long.MAX_VALUE));
		parser.parse(trace);

		float fitness = parser.getNumParsedElements();
		//float fitness = trace.size();
		fitness -= parser.getNumExtraTokensLeftBehind();
		fitness -= parser.getNumMissingTokens();
		fitness /= trace.size();
		return fitness;
	}

	/**
	 * The current net.
	 */
	private HeuristicsNet net;

	/**
	 * The represented XLog.
	 */
	private XLog log;

	/**
	 * The settings for HeuristicsMiner.
	 */
	private HeuristicsMinerSettings settings;

	/**
	 * Create a new ProcessModel for the given XLog and with the given
	 * HeuristicsMiner settings. This will immediately call HeuristicsMiner to
	 * calculate the process model.
	 * 
	 * @param log
	 *            The XLog to be represented.
	 * @param settings
	 *            The settings for HeuristicsMiner.
	 */
	public ProcessModel(XLog log, HeuristicsMinerSettings settings) {
		this.log = log;
		this.settings = settings;

		// pas de settings aan in functie van input gebruiker en frequentie van de 
		// log traces size/size+1 - 0.01
		
		// For small logs, the dependency thresholds need to be adapted
		
		HeuristicsMinerSettings adaptedSettings = settings;
		
		double threshold = ((double)log.size() / ((double)log.size()+1));
		double minthreshold = threshold - 0.01;
		
		//System.out.println("Log size is "+log.size()+" treshold set to "+threshold);
		
		
		
		if ( threshold < adaptedSettings.getDependencyThreshold()) {
			adaptedSettings.setDependencyThreshold(minthreshold);
		}
		
		if (threshold < adaptedSettings.getL1lThreshold()){
			adaptedSettings.setL1lThreshold(minthreshold);
		}
		
		if (threshold < adaptedSettings.getL2lThreshold()){
			adaptedSettings.setL2lThreshold(minthreshold);
		}
		
		//System.out.println(adaptedSettings.toString());
		
		HeuristicsMinerNoUI miner = new HeuristicsMinerNoUI(log, adaptedSettings);
		net = miner.mine();
		//System.out.println("Fitness: " + net.getFitness());
	}

	/**
	 * Create a new ProcessModel by taking the log of this model and adding the
	 * given log to it. The current ProcessModel is left unchanged.
	 * 
	 * @param log
	 *            The log that should be added to the current log in the new
	 *            object.
	 * @return A new ProcessModel, representing the given XLog plus the current
	 *         XLog.
	 */
	public ProcessModel add(XLog log) {
		XLog new_log = new XLogImpl(log.getAttributes());
		new_log.addAll(log);
		new_log.addAll(this.log);
		return new ProcessModel(new_log, settings);
	}

	/**
	 * Returns the ICS-fitness of the represented process model.
	 */
	public float getFitness() {
		return (float) net.getFitness();
	}

	/**
	 * Return the underlying process model in the form of a HeuristicsNet.
	 */
	public HeuristicsNet getHeuristicsNet() {
		return net;
	}

	/**
	 * Calculate the fitness of the given trace in the current model, without
	 * changing the model.
	 */
	public float traceFitness(XTrace trace) {
		return traceFitness(trace, net.copy());
	}
}
