package org.processmining.plugins.actitrac.types;

import java.util.ArrayList;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.plugins.actitrac.algorithm.ProcessModel;
import org.processmining.plugins.actitrac.utils.XEventNameAndTransitionClassifier;
import org.processmining.plugins.heuristicsnet.miner.heuristics.miner.fitness.ImprovedContinuousSemantics;


/**
 * This class represents a single cluster. A cluster is a representation of an
 * XLog with an underlying ProcessModel and a name.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-07-29
 */
public class Cluster {

	/**
	 * An XLog with the traces that are part of this cluster and used to
	 * calculate the ProcessModel.
	 */
	private XLog internalLog;

	/**
	 * An XLog with the traces that are part of this cluster but not used to
	 * calculate the ProcessModel.
	 */
	private XLog externalLog;

	/**
	 * The underlying process model.
	 */
	private ProcessModel model;

	/**
	 * The name of this cluster.
	 */
	private String name;

	/**
	 * An XEventClassifier that will be used to compare XEvents for equality.
	 */
	private static XEventClassifier classifier = new XEventNameAndTransitionClassifier();

	/**
	 * Checks whether or not the two given traces are similar. Two traces are
	 * considered to be distinct when the i-th event in the first process
	 * instance is not the same as the i-th event in the second process
	 * instance. Events are defined as being the same when their concept:name
	 * and lifecycle:transition both are the same.
	 * 
	 * @param one
	 *            The first XTrace to compare.
	 * @param two
	 *            The second XTrace to compare.
	 * @return True if and only if the two traces are similar. False otherwise.
	 */
	private static boolean tracesSimilar(XTrace one, XTrace two) {
		if (one.size() != two.size())
			return false;
		for (int i = 0; i < one.size(); i++)
			if (!classifier.sameEventClass(one.get(i), two.get(i)))
				return false;
		return true;
	}

	/**
	 * Create a new Cluster, with the given internal log, model and name.
	 * 
	 * @param log
	 *            The internal log for this cluster.
	 * @param model
	 *            The process model (this should be a model calculated based on
	 *            the given log).
	 * @param name
	 *            The name of this cluster.
	 */
	public Cluster(XLog log, ProcessModel model, String name) {
		this.internalLog = log;
		this.model = model;
		this.externalLog = new XLogImpl(log.getAttributes());
		this.name = name;
	}

	/**
	 * Add the traces in the given XLog to the external traces.
	 */
	public void addExternal(XLog log) {
		externalLog.addAll(log);
		for (XTrace trace : log) {
			ProcessModel.traceFitness(trace, model.getHeuristicsNet());
		}
	}

	/**
	 * Return the amount of DPI in the external log.
	 */
	public int getExternalGroupCount() {
		ArrayList<XTrace> groups = new ArrayList<XTrace>();
		for (XTrace trace : getExternalLog()) {
			boolean found = false;
			for (XTrace group : groups) {
				if (tracesSimilar(trace, group)) {
					found = true;
				}
			}
			if (!found) {
				groups.add(trace);
			}
		}
		return groups.size();
	}

	/**
	 * Return a copy of the external log of this cluster. Traces in the external
	 * log of a cluster have not been used to calculate the process model of
	 * that cluster.
	 */
	public XLog getExternalLog() {
		return (XLog) externalLog.clone();
	}

	/**
	 * Return the total ICS-fitness of this Cluster, when evaluated for this
	 * cluster's process model.
	 * 
	 * @return The ICS-fitness of this cluster.
	 */
	public float getFitness() {
		if (externalLog.size() == 0)
			return (float) model.getHeuristicsNet().getFitness();
		XLog log = new XLogImpl(internalLog.getAttributes());
		log.addAll(internalLog);
		log.addAll(externalLog);
		ImprovedContinuousSemantics ics = new ImprovedContinuousSemantics(
				XLogInfoFactory.createLogInfo(log));
		HeuristicsNet[] nets = ics.calculate(new HeuristicsNet[] { model
				.getHeuristicsNet().copy() });
		return (float) nets[0].getFitness();
	}

	/**
	 * Return the amount of DPI in the internal log.
	 */
	public int getInternalGroupCount() {
		ArrayList<XTrace> groups = new ArrayList<XTrace>();
		for (XTrace trace : getInternalLog()) {
			boolean found = false;
			for (XTrace group : groups) {
				if (tracesSimilar(trace, group)) {
					found = true;
				}
			}
			if (!found) {
				groups.add(trace);
			}
		}
		return groups.size();
	}

	/**
	 * Return a copy of the internal log of this Cluster. Traces in the internal
	 * log have been used to build the process model for this cluster.
	 */
	public XLog getInternalLog() {
		return (XLog) internalLog.clone();
	}

	/**
	 * Return the name of this cluster.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Return the ProcessModel that represents this cluster.
	 */
	public ProcessModel getProcessModel() {
		return model;
	}
}
