package org.processmining.plugins.actitrac.types;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.plugins.actitrac.utils.XEventNameAndTransitionClassifier;


/**
 * This class represents an XLog of which the traces have been grouped. XTraces
 * are grouped together when the classes of their XEvents are equal and in the
 * same order. To determine the class of an XEvent the Classifier
 * {@link XEventNameAndTransitionClassifier} is used.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-07-29
 */
public class GroupedXLog implements Cloneable {

	/**
	 * The used classifier-object.
	 */
	private XEventClassifier classifier = new XEventNameAndTransitionClassifier();

	/**
	 * The groups of the XTraces. These should be sorted according to the size
	 * of their sub-lists. Bigger lists come first.
	 */
	private List<List<XTrace>> groups;

	/**
	 * The XAttributeMap of the given XLog.
	 */
	private XAttributeMap attributeMap;

	/**
	 * The total amount of traces in this log.
	 */
	private int nbTraces = 0;

	/**
	 * The name of the represented XLog.
	 */
	private String name;

	/**
	 * Empty constructor, this should only be used by the clone-method.
	 */
	private GroupedXLog() {
	}

	/**
	 * Create a new GroupedXLog based on an XLog. This will group and sort the
	 * XTraces in the given XLog.
	 * 
	 * @param log
	 *            The XLog that should be represented by this GroupedXLog.
	 */
	public GroupedXLog(XLog log) {
		attributeMap = log.getAttributes();
		groups = new ArrayList<List<XTrace>>();
		{
			name = log.getAttributes().get("concept:name").toString();
			int length = name.lastIndexOf('.');
			if (length >= 0) {
				if (name.subSequence(length, name.length()).equals(".gz")) {
					length = name.lastIndexOf('.', length - 1);
				}
				if (length != 0) {
					name = name.substring(0, length);
				}
			}
		}
		for (XTrace trace : log) {
			nbTraces++;
			int index = -1;
			for (int i = 0; i < groups.size(); i++) {
				if (tracesSimilar(trace, groups.get(i).get(0))) {
					index = i;
					break;
				}
			}
			if (index == -1) {
				index = groups.size();
				groups.add(new ArrayList<XTrace>());
			}
			groups.get(index).add(trace);
		}
		Collections.sort(groups, new Comparator<List<XTrace>>() {

			@Override
			public int compare(List<XTrace> a, List<XTrace> b) {
				return b.size() - a.size();
			}
		});
	}

	@Override
	public Object clone() {
		GroupedXLog output = new GroupedXLog();
		output.classifier = new XEventNameAndTransitionClassifier();
		output.groups = new ArrayList<List<XTrace>>();
		for (int i = 0; i < groups.size(); i++) {
			XTrace[] sub_array = groups.get(i).toArray(new XTrace[0]);
			output.groups.add(new ArrayList<XTrace>());
			for (int j = 0; j < sub_array.length; j++) {
				output.groups.get(i).add(sub_array[j]);
			}
		}
		output.attributeMap = (XAttributeMap) attributeMap.clone();
		output.nbTraces = nbTraces;
		output.name = name;
		return output;
	}

	/**
	 * Get the index-th group of this log. Groups are sorted by size, with the
	 * largest groups first. All XTrace objects in this group are similar.
	 * 
	 * @param index
	 *            The index of the group.
	 * @return A list of XTrace objects that are part of this group.
	 */
	public List<XTrace> get(int index) {
		return groups.get(index);
	}

	/**
	 * Return a map containing the attributes for this log.
	 */
	public XAttributeMap getAttributeMap() {
		return attributeMap;
	}

	/**
	 * Creates and returns the entire XLog that is represented by this
	 * GroupedXLog.
	 * 
	 * @return An XLog containing all of the traces in this GroupedXLog.
	 */
	public XLog getLog() {
		XLogImpl log = new XLogImpl(attributeMap);
		for (int i = 0; i < groups.size(); i++)
			log.addAll(groups.get(i));
		return log;
	}

	/**
	 * Creates and returns an XLog based on the groups denoted by the given
	 * indices.
	 * 
	 * @see #get(int)
	 * @param indices
	 *            The indices of the groups that ought to be part of the
	 *            generated XLog.
	 * @return An XLog containing all of the traces in the groups denoted by the
	 *         given indices.
	 */
	public XLog getLog(Collection<Integer> indices) {
		XLogImpl log = new XLogImpl(attributeMap);
		for (Integer i : indices)
			log.addAll(groups.get(i));
		return log;
	}

	/**
	 * Creates and returns an XLog based on the groups denoted by the given
	 * indices.
	 * 
	 * @see #get(int)
	 * @param indices
	 *            The indices of the groups that ought to be part of the
	 *            generated XLog.
	 * @return An XLog containing all of the traces in the groups denoted by the
	 *         given indices.
	 */
	public XLog getLog(int... indices) {
		XLogImpl log = new XLogImpl(attributeMap);
		for (int i = 0; i < indices.length; i++)
			log.addAll(groups.get(indices[i]));
		return log;
	}

	/**
	 * The name of the represented XLog.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Retuns the total amount of traces in this log.
	 */
	public int getNbTraces() {
		return nbTraces;
	}

	/**
	 * Remove a group from this XLog.
	 * 
	 * @param index
	 *            The index of the group to be removed.
	 */
	public void remove(int index) {
		nbTraces -= groups.get(index).size();
		groups.remove(index);
	}

	/**
	 * Remove the group from this XLog with traces that are similar to the given
	 * trace.
	 * 
	 * @param trace
	 *            The trace whose similar group has to be removed.
	 */
	public void remove(XTrace trace) throws IllegalArgumentException {
		for (int i = 0; i < groups.size(); i++) {
			if (tracesSimilar(trace, groups.get(i).get(0))) {
				nbTraces -= groups.get(i).size();
				groups.remove(i);
				return;
			}
		}
		throw new IllegalArgumentException("No similar group found");
	}

	/**
	 * Returns the amount of groups of which this log consists.
	 */
	public int size() {
		return groups.size();
	}

	/**
	 * Determine whether or not the two given XTrace objects are similar. XTrace
	 * objects are considered to be similar when the classes of their XEvents
	 * are equal and in the same order. To determine the class of an XEvent the
	 * Classifier {@link XEventNameAndTransitionClassifier} is used.
	 * 
	 * @param one
	 * @param two
	 * @return True if the given XTrace objects are similar, false otherwise.
	 */
	private boolean tracesSimilar(XTrace one, XTrace two) {
		if (one.size() != two.size())
			return false;
		for (int i = 0; i < one.size(); i++) {
			if (!classifier.sameEventClass(one.get(i), two.get(i)))
				return false;
		}
		return true;
	}
}
