package org.processmining.plugins.actitrac.ui;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager2;
import java.awt.Point;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.border.Border;

import org.processmining.plugins.actitrac.utils.Line;


/**
 * This implementation of a LayoutManager will lay out its components in the
 * form of a binary tree. The root of this tree is located in the top-left
 * corner. The right child of each node, if such child exists, is positioned
 * immediately right of the node whose child it is, including the entire
 * sub-tree of which this child node is the root. The left child of this parent
 * node and its entire sub-tree of which it is the root is then laid out
 * immediately below the lowest point that the sub-tree of the right child
 * reaches. Every node in the tree will have exactly the same size, but no
 * resizing will happen to the preferred size of any component.
 * 
 * Not every position of the tree has to be filled, neither does every node need
 * a parent. However, the behaviour of the method {{@link #getLines()} is
 * undefined in such case. There is also no strict checking of loops or
 * components that have been added multiple times without being removed. In this
 * case, too, the behaviour is undefined.
 * 
 * The constraints which this layout manager will accept are strings defining
 * the node in which this component is located. These constraint strings can
 * contain any character, but only the characters 'l' and 'r' are taken into
 * account. The root of the tree should be given the constraint "". In the logic
 * of these constraint strings, this means that no branches have been taken. The
 * direct left child of any node should be given the string formed by taking the
 * parent's constraint string and adding 'l'. The same is true for the direct
 * right child, in which case 'r' has to be added to the parent's constraint
 * string.
 * 
 * Finally, this class also includes a way of drawing the lines that connect the
 * nodes of the tree. See {@link #getLines()} for more information on how to do
 * this.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-01
 */
public class BinaryTreeLayout implements LayoutManager2 {

	/**
	 * A representation of the tree that the layout manager will use to remember
	 * the component's positions.
	 */
	private class TreeNode {
		TreeNode left = null;
		TreeNode right = null;
		TreeNode parent = null;
		Component data = null;
	}

	/**
	 * The root of the internal tree.
	 */
	private TreeNode root = new TreeNode();

	/**
	 * A representation of the cells of the layout.
	 */
	private ArrayList<ArrayList<Component>> components = null;

	/**
	 * The width of the widest cell.
	 */
	private int maxCellWidth = 0;

	/**
	 * The height of the highest cell.
	 */
	private int maxCellHeight = 0;

	/**
	 * The depth of the deepest node.
	 */
	private int maxTreeDepth = 0;

	/**
	 * The amount of space between any two node cells on the X-axis.
	 */
	private int spacingX = 30;

	/**
	 * The amount of space between any two node cells on the Y-axis.
	 */
	private int spacingY = 20;

	@Override
	public void addLayoutComponent(Component comp, Object arg1) {
		if (!(arg1 instanceof String))
			throw new IllegalArgumentException(
					"Only strings are accepted as constraints");
		String constraints = (String) arg1;
		TreeNode position = root;
		for (int i = 0; i < constraints.length(); i++) {
			if (constraints.charAt(i) == 'r') {
				if (position.right == null) {
					position.right = new TreeNode();
					position.right.parent = position;
				}
				position = position.right;
			} else if (constraints.charAt(i) == 'l') {
				if (position.left == null) {
					position.left = new TreeNode();
					position.left.parent = position;
				}
				position = position.left;
			}
		}
		position.data = comp;
		invalidateLayout();
	}

	@Override
	public void addLayoutComponent(String name, Component comp) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Make sure that the components have been laid out and if not, do so.
	 */
	private void ensureComponentsLaidOut() {
		if (components == null)
			layout();
	}

	@Override
	public float getLayoutAlignmentX(Container target) {
		return 0.5f;
	}

	@Override
	public float getLayoutAlignmentY(Container target) {
		return 0.5f;
	}

	/**
	 * Generate and return the lines that make up the relations of this tree.
	 * For any node that has a child, the lines will converge to the point in
	 * the middle of this component. For any component without children, the
	 * lines will converge to the point in the middle of the component on the
	 * Y-axis and at the left side of this component.
	 * 
	 * These lines can, for example, be used by the parent component. By drawing
	 * these lines in the paint- or paintComponent-method, they could be
	 * visualised along with the components that make up the tree.
	 * 
	 * @return
	 *         A list of Line objects that make up the parent-child relations of
	 *         the tree which this BinaryTreeLayout represents.
	 */
	public List<Line> getLines() {
		ensureComponentsLaidOut();
		List<Line> lines = new ArrayList<Line>();
		LinkedList<TreeNode> stack = new LinkedList<TreeNode>();
		stack.add(root);
		while (!stack.isEmpty()) {
			TreeNode current = stack.removeFirst();
			Point start = null;
			if (current.data != null) {
				start = new Point();
				start.x = current.data.getX() + current.data.getWidth() / 2;
				start.y = current.data.getY() + current.data.getHeight() / 2;
			}
			if (current.right != null) {
				stack.add(current.right);
				if (current.right.data != null) {
					Point end = new Point();
					end.x = current.right.data.getX();
					end.x += current.right.data.getWidth() / 2;
					end.y = current.right.data.getY();
					end.y += current.right.data.getHeight() / 2;
					if (current.right.right == null
							&& current.right.left == null)
						end.x -= current.right.data.getPreferredSize().width / 2;
					lines.add(new Line(start, end));
				}
			}
			if (current.left != null) {
				stack.add(current.left);
				if (current.left.data != null) {
					Point end = new Point();
					end.x = current.left.data.getX();
					end.x += current.left.data.getWidth() / 2;
					end.y = current.left.data.getY();
					end.y += current.left.data.getHeight() / 2;
					if (current.left.right == null && current.left.left == null)
						end.x -= current.left.data.getPreferredSize().width / 2;
					lines.add(new Line(start, end));
				}
			}
		}
		return lines;
	}

	/**
	 * Return the amount of space that is left between any two components on the
	 * X-axis.
	 */
	public int getSpacingX() {
		return spacingX;
	}

	/**
	 * Return the amount of space that is left between any two components on the
	 * Y-axis.
	 * 
	 * @return
	 */
	public int getSpacingY() {
		return spacingY;
	}

	/**
	 * Invalidates the layout, so that any subsequent call that requires
	 * information on the way components are laid out will have to recalculate
	 * the actual positioning of the components.
	 */
	private void invalidateLayout() {
		components = null;
		maxCellWidth = 0;
		maxCellHeight = 0;
		maxTreeDepth = 0;
	}

	@Override
	public void invalidateLayout(Container target) {
		invalidateLayout();
	}

	/**
	 * Calculate in which cell each component should be placed.
	 */
	private void layout() {
		components = new ArrayList<ArrayList<Component>>();
		layout(0, 0, root);
	}

	/**
	 * Lay out the given TreeNode and the sub-tree whose root is this TreeNode.
	 * The amount of space which this method can use is not limited on the
	 * X-axis or Y-axis. However, no component can be placed at a lower X-value
	 * than the given X-value and no component can be placed at a lower Y-value
	 * than the given Y-value.
	 * 
	 * @param x
	 *            The minimal X-value at which any component of the sub-tree of
	 *            which 'current' is the root can be placed.
	 * @param y
	 *            The minimal Y-value at which any component of the sub-tree of
	 *            which 'current' is the root can be placed.
	 * @param current
	 *            The current TreeNode that should be laid out.
	 * @return The lowest Y-value that has not been touched by this sub-tree.
	 */
	private int layout(int x, int y, TreeNode current) {
		if (x > maxTreeDepth)
			maxTreeDepth = x;
		// Lay out yourself on the current position
		while (components.size() <= y)
			components.add(new ArrayList<Component>());
		while (components.get(y).size() <= x)
			components.get(y).add(null);
		if (current.data != null) {
			components.get(y).set(x, current.data);
			Dimension d = current.data.getPreferredSize();
			if (d.height > maxCellHeight)
				maxCellHeight = d.height;
			if (d.width > maxCellWidth)
				maxCellWidth = d.width;
		}
		// If you have a right child, lay it out
		// This returns the lowest y-value that he has used to lay out himself +
		// 1.
		if (current.right != null) {
			y = layout(x + 1, y, current.right);
		}
		// If you have a left child, lay it out
		// Start doing so at an y-value that has not yet been used.
		if (current.left != null) {
			y = layout(x + 1, y, current.left);
		}
		// Return the lowest y-value that has not yet been used.
		return y + 1;
	}

	@Override
	public void layoutContainer(Container parent) {
		ensureComponentsLaidOut();
		Insets insets;
		try {
			JComponent jParent = (JComponent) parent;
			Border b = jParent.getBorder();
			insets = b.getBorderInsets(parent);
		} catch (ClassCastException e) {
			insets = new Insets(0, 0, 0, 0);
		}
		for (int y = 0; y < components.size(); y++) {
			for (int x = 0; x < components.get(y).size(); x++) {
				Component c = components.get(y).get(x);
				if (c != null) {
					Dimension d = c.getPreferredSize();
					c.setBounds(insets.left + x * (maxCellWidth + spacingX)
							+ (maxCellWidth - d.width) / 2, insets.top + y
							* (maxCellHeight + spacingY)
							+ (maxCellHeight - d.height) / 2, d.width, d.height);
				}
			}
		}
	}

	@Override
	public Dimension maximumLayoutSize(Container target) {
		return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	@Override
	public Dimension minimumLayoutSize(Container parent) {
		return preferredLayoutSize(null);
	}

	@Override
	public Dimension preferredLayoutSize(Container parent) {
		ensureComponentsLaidOut();
		Insets insets;
		try {
			JComponent jParent = (JComponent) parent;
			Border b = jParent.getBorder();
			insets = b.getBorderInsets(parent);
		} catch (ClassCastException e) {
			insets = new Insets(0, 0, 0, 0);
		}
		Dimension d = new Dimension();
		d.height = components.size() * (maxCellHeight + spacingY) + insets.top
				+ insets.bottom;
		d.width = (maxTreeDepth + 1) * (maxCellWidth + spacingX) + insets.left
				+ insets.right;
		return d;
	}

	@Override
	public void removeLayoutComponent(Component comp) {
		LinkedList<TreeNode> stack = new LinkedList<TreeNode>();
		stack.add(root);
		while (!stack.isEmpty()) {
			TreeNode current = stack.removeFirst();
			if (current.data == comp) {
				current.data = null;
				invalidateLayout();
				if (current.left == null && current.right == null
						&& current.parent != null) {
					if (current.parent.right == current)
						current.parent.right = null;
					else
						current.parent.left = null;
				}
				return;
			} else {
				if (current.left != null)
					stack.add(current.left);
				if (current.right != null)
					stack.add(current.right);
			}
		}
	}

	/**
	 * Set the amount of space that is left between any two components on the
	 * X-axis to the given value.
	 */
	public void setSpacingX(int spacingX) {
		this.spacingX = spacingX;
	}

	/**
	 * Set the amount of space that is left between any two components on the
	 * Y-axis to the given value.
	 */
	public void setSpacingY(int spacingY) {
		this.spacingY = spacingY;
	}
}
