package org.processmining.plugins.actitrac.ui;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.utils.ATCParameters;
import org.processmining.plugins.actitrac.utils.ClustererEventListener;


/**
 * This implementation of a ClustererEventListener forms a bridge between a
 * Clusterer object and a UIPluginContext. It shall reflect the progress of the
 * Clusterer in the UIPluginContext and correctly implement the UI's
 * cancel-operation.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-05
 */
public class ClustererEventListenerUI implements ClustererEventListener {

	/**
	 * The UIPluginContext to interface to.
	 */
	private UIPluginContext context;

	/**
	 * Whether the Clusterer shall stop when a certain amount of clusters has
	 * been created. The only other option is to stop when a certain percentage
	 * of traces has been clustered.
	 */
	private boolean stopByNbGroups;

	/**
	 * The total amount of traces that the Clusterer has processed.
	 */
	private int totalTracesDone = 0;

	/**
	 * The number of the Cluster that is currently being created.
	 */
	private int clusterNb = 1;

	/**
	 * The amount of work (for use with progress bars) that one unit of work
	 * equals to. This unit of work can be a cluster or a trace.
	 */
	private float workPerUnit;

	/**
	 * The amount of nonfitting traces that has been assigned to a cluster.
	 */
	private int nonfitting = 0;

	/**
	 * Create a new ClustererEventListenerUI for the given UIPluginContext, XLog
	 * and ATCParameters.
	 */
	public ClustererEventListenerUI(UIPluginContext context, XLog log,
			ATCParameters parameters) {
		this.context = context;
		stopByNbGroups = parameters.getStopByNbGroups();
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(Integer.MAX_VALUE);
		if (stopByNbGroups) {
			int targetGroups = parameters.getStopNbGroups();
			context.getProgress().setValue(0);
			workPerUnit = Integer.MAX_VALUE / 2f / targetGroups;
		} else {
			int totalTraces = log.size();
			context.getProgress().setValue(0);
			workPerUnit = Integer.MAX_VALUE / 2f / totalTraces
					* parameters.getStopPercentage() / 100f;
		}
		if (parameters.getSeparateNonFitting())
			workPerUnit *= 2;
	}

	@Override
	public void cancel() {
		context.log("Cancelled");
		context.getProgress().setValue(Integer.MAX_VALUE);
	}

	@Override
	public void clusterCreated(Cluster c) {
		String log = "Cluster #" + clusterNb + ": ";
		log += c.getInternalLog().size() + " internal traces; ";
		log += c.getInternalGroupCount() + " internal groups; ";
		log += c.getProcessModel().getFitness() + " fitness;";
		context.log(log);

		totalTracesDone += c.getInternalLog().size();
		if (stopByNbGroups) {
			context.getProgress().setValue((int) (clusterNb * workPerUnit));
		} else {
			context.getProgress().setValue(
					(int) (totalTracesDone * workPerUnit));
		}
		clusterNb++;
	}

	@Override
	public void end() {
		context.log("Finished");
	}

	@Override
	public boolean minimalClusterSizeImpossible() {
		// int result = JOptionPane
		// .showConfirmDialog(
		// null,
		// "The specified minimal cluster size cannot be reached under the given parameters. Would you like to continue?",
		// "Warning", JOptionPane.YES_NO_OPTION);
		// if (result == JOptionPane.YES_OPTION)
		// return true;
		// else
		// return false;
		context.log("Cluster #" + clusterNb
				+ ": minimal cluster size could not be reached.");
		return true;
	}

	@Override
	public void nonfittingAssigned() {
		nonfitting++;
		context.getProgress().setValue(
				(int) (Integer.MAX_VALUE / 2 + nonfitting * workPerUnit));
	}

	@Override
	public boolean shouldCancel() {
		return context.getProgress().isCancelled();
	}

	@Override
	public void start() {
		context.getProgress().setCaption("Calculating clusters");
		context.log("Started calculations");
	}

	@Override
	public void targetReached(int left) {
		context.log("Target reached, adding " + left
				+ " groups of non-fitting traces");
		workPerUnit = Integer.MAX_VALUE / 2f / left;
	}

	@Override
	public void importing() {
		context.getProgress().setCaption("Importing data");
		context.log("Importing data");
	}
}
