package org.processmining.plugins.actitrac.ui;

import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JPanel;

import org.processmining.plugins.actitrac.ActiTraCClustering;
import org.processmining.plugins.actitrac.types.ClusterTreeInnerNode;
import org.processmining.plugins.actitrac.types.ClusterTreeLeaf;
import org.processmining.plugins.actitrac.types.ClusterTreeNode;
import org.processmining.plugins.actitrac.types.ClusterTreeRoot;
import org.processmining.plugins.actitrac.ui.clustertree.JClusterBox;
import org.processmining.plugins.actitrac.ui.clustertree.JClusterSplitBox;
import org.processmining.plugins.actitrac.ui.clustertree.JRootBox;
import org.processmining.plugins.actitrac.utils.Line;


/**
 * This graphical component will create a view of an entire ActiTraC-clustering.
 * It will lay out all of the clusters of the clustering in a binary tree. When
 * this component is clicked, it will also notify the assigned JClusteringSidePane
 * of this.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-01
 */
public class JClusteringPane extends JPanel {

	private static final long serialVersionUID = -5805795366175833932L;

	/**
	 * The LayoutManager which we are going to use.
	 */
	private BinaryTreeLayout layout;

	/**
	 * The JClusteringSidePane to which we should report selection updates.
	 */
	private JClusteringSidePane clusterinfo;

	/**
	 * This will notify clusterinfo that nothing is selected when this component
	 * is clicked.
	 */
	MouseListener jClusteringPaneMouseListener = new MouseListener() {

		@Override
		public void mouseClicked(MouseEvent e) {
			clusterinfo.setSelectedBox(null);
			clusterinfo.showClusterInfo(null);
			e.consume();
		}

		@Override
		public void mouseEntered(MouseEvent arg0) {
		}

		@Override
		public void mouseExited(MouseEvent arg0) {
		}

		@Override
		public void mousePressed(MouseEvent arg0) {
		}

		@Override
		public void mouseReleased(MouseEvent arg0) {
		}
	};

	/**
	 * Create a new JClusteringPane for the given ActiTraCClustering. This new
	 * JClusteringPane will notify the given JClusteringSidePane when it is
	 * clicked. Creating a JClusteringPane may result in the creation of other
	 * graphical components (children of this newly created JClusteringPane),
	 * which may too contact the given JClusteringSidePane when deemed appropriate.
	 */
	public JClusteringPane(ActiTraCClustering clustering,
			JClusteringSidePane clusterinfo) {
		// A small class whose only purpose in life is to remember where we
		// found a certain node.
		class StackItem {
			String position;
			ClusterTreeNode node;

			StackItem(String position, ClusterTreeNode node) {
				this.position = position;
				this.node = node;
			}
		}

		this.clusterinfo = clusterinfo;
		this.setOpaque(false);
		this.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 220));
		layout = new BinaryTreeLayout();
		this.setLayout(layout);

		// Go over every node in the ClusterTree. Build the
		// BinaryTreeLayout-constraints while doing this.
		// Also, every node has some sort of graphical representation.
		LinkedList<StackItem> stack = new LinkedList<StackItem>();
		stack.add(new StackItem("", clustering.getTree().getRoot()));
		while (!stack.isEmpty()) {
			StackItem item = stack.removeFirst();
			if (item.node instanceof ClusterTreeInnerNode) {
				stack.add(new StackItem(item.position + "r", item.node
						.getRight()));
				stack.add(new StackItem(item.position + "l", item.node
						.getLeft()));
				if (item.node instanceof ClusterTreeRoot) {
					ClusterTreeRoot root = (ClusterTreeRoot) item.node;
					this.add(
							new JRootBox(root.getPseudoCluster(), clustering
									.getLog(), clusterinfo), item.position);
				} else {
					this.add(new JClusterSplitBox(), item.position);
				}
			} else if (item.node instanceof ClusterTreeLeaf) {
				this.add(
						new JClusterBox(((ClusterTreeLeaf) item.node)
								.getCluster(), clustering.getLog(), clusterinfo),
						item.position);
			}
		}
		this.addMouseListener(jClusteringPaneMouseListener);
	}

	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setStroke(new BasicStroke(2));
		List<Line> lines = layout.getLines();
		for (Line line : lines) {
			g2.drawLine(line.start.x, line.start.y, line.end.x, line.end.y);
		}
	}
}
