package org.processmining.plugins.actitrac.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.plugins.actitrac.ActiTraCClustering;
import org.processmining.plugins.actitrac.algorithm.ClusterStatistics;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.ui.clustertree.JClusterBox;
import org.processmining.plugins.heuristicsnet.visualizer.HeuristicsNetAnnotatedVisualization;


import com.fluxicon.slickerbox.factory.SlickerFactory;

/**
 * This graphical component will perform a number of functions.
 * 
 * Firstly, it will maintain a concept of selection. At any time, one or none
 * JClusterBox can be selected. When a JClusterBox is selected, this JClusterBox
 * will need to be given a distinct style.
 * 
 * Secondly, it will be able to display detailed information of a single
 * Cluster. When the detailed information of a Cluster is show, it will also be
 * able to export and/or view some of the data related to that cluster.
 * 
 * Lastly, at all times it should be able to export some information of each
 * cluster without much effort for the user.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-01
 */
public class JClusteringSidePane extends JPanel {

	/**
	 * This class will export the complete XLog of each Cluster to ProM.
	 */
	private class AllLogsExportAction implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			for (Cluster cluster : clustering.getTree()) {
				XLog log = cluster.getInternalLog();
				log.addAll(cluster.getExternalLog());
				context.getProvidedObjectManager().createProvidedObject(
						cluster.getName(), log, XLog.class, context);
			}
		}
	}

	/**
	 * This class will be able to export the HeuristicsNet of each Cluster to
	 * ProM.
	 */
	private class AllNetsExportAction implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			for (Cluster cluster : clustering.getTree()) {
				HeuristicsNet net = cluster.getProcessModel()
						.getHeuristicsNet();
				context.getProvidedObjectManager().createProvidedObject(
						cluster.getName(), net, HeuristicsNet.class, context);
			}
		}
	}

	/**
	 * This class will be able to export the complete XLog of a certain Cluster
	 * to ProM.
	 */
	private class ExportLogAction implements ActionListener {

		private Cluster cluster;

		public ExportLogAction(Cluster cluster) {
			this.cluster = cluster;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			XLog log = cluster.getInternalLog();
			log.addAll(cluster.getExternalLog());
			String name = cluster.getName();
			context.getProvidedObjectManager().createProvidedObject(name, log,
					XLog.class, context);
		}
	}

	/**
	 * This class will be able to export the HeuristicsNet of a certain Cluster
	 * to ProM.
	 */
	private class ExportNetAction implements ActionListener {

		private Cluster cluster;

		public ExportNetAction(Cluster cluster) {
			this.cluster = cluster;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			HeuristicsNet net = cluster.getProcessModel().getHeuristicsNet();
			String name = cluster.getName();
			context.getProvidedObjectManager().createProvidedObject(name, net,
					HeuristicsNet.class, context);
		}
	}

	/**
	 * This class will be able to create a view of the HeuristicsNet of a
	 * certain Cluster.
	 */
	private class ViewNetAction implements ActionListener {

		private Cluster cluster;

		public ViewNetAction(Cluster cluster) {
			this.cluster = cluster;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			HeuristicsNet net = cluster.getProcessModel().getHeuristicsNet();
			String name = cluster.getName();
			JFrame frame = new JFrame(name);
			PluginContext subcontext = context
					.createChildContext("HeuristicsNetVisualization");
			JComponent visualization = HeuristicsNetAnnotatedVisualization
					.visualize(subcontext, net);
			frame.add(visualization);
			Toolkit kit = Toolkit.getDefaultToolkit();
			Dimension size = kit.getScreenSize();
			size.width = size.width * 2 / 3;
			size.height = size.height * 2 / 3;
			frame.setSize(size);
			frame.setVisible(true);
		}
	}

	private static final long serialVersionUID = -220929134473510430L;

	/**
	 * The foreground colour of this object. Mostly used for the text.
	 */
	private Color foreground = new Color(240, 240, 240);

	/**
	 * The top part of the JClusteringSidePane. Contains anything that is
	 * related to a specific Cluster.
	 */
	private JPanel top_panel;

	/**
	 * The bottom part of the JClusteringSidePane. Contains anything that is
	 * related to the entire Clustering.
	 */
	private JPanel bottom_panel;

	/**
	 * The context in which we currently work.
	 */
	private PluginContext context;

	/**
	 * The ActiTraCClustering which is currently being displayed.
	 */
	private ActiTraCClustering clustering;

	/**
	 * The JClusterBox that is currently selected, or null if no JClusterBox is
	 * selected.
	 */
	private JClusterBox selected = null;

	/**
	 * Will consume everything that lands on the panel and isn't consumed by
	 * anything else yet. This is because we are working with panels that are
	 * not entirely opaque.
	 */
	private MouseListener jClusterInfoPaneMouseListener = new MouseListener() {

		@Override
		public void mouseClicked(MouseEvent arg0) {
			arg0.consume();
		}

		@Override
		public void mouseEntered(MouseEvent arg0) {
		}

		@Override
		public void mouseExited(MouseEvent arg0) {
		}

		@Override
		public void mousePressed(MouseEvent arg0) {
		}

		@Override
		public void mouseReleased(MouseEvent arg0) {
		}
	};

	/**
	 * Create a new JClusteringSidePane for the given clustering, in the given
	 * context.
	 */
	public JClusteringSidePane(PluginContext context,
			ActiTraCClustering clustering) {
		this.context = context;
		this.clustering = clustering;

		this.setOpaque(false);
		this.setBorder(BorderFactory.createEmptyBorder(20, 35, 20, 5));
		this.setLayout(new BorderLayout());
		this.setForeground(new Color(240, 240, 240));
		this.setPreferredSize(new Dimension(280, 250));

		top_panel = new JPanel();
		bottom_panel = new JPanel();

		initTopPanel();
		initBottomPanel();

		this.add(top_panel, BorderLayout.PAGE_START);
		this.add(bottom_panel, BorderLayout.PAGE_END);

		this.addMouseListener(jClusterInfoPaneMouseListener);
	}

	/**
	 * Initialise the bottom panel.
	 */
	private void initBottomPanel() {
		SlickerFactory slicker = SlickerFactory.instance();
		bottom_panel.setLayout(new BoxLayout(bottom_panel, BoxLayout.Y_AXIS));
		bottom_panel.setOpaque(false);
		JButton allLogsToProM = slicker.createButton("Logs to ProM");
		JButton allNetsToProM = slicker.createButton("HeuristicsNets to ProM");
		JLabel exportLabel = newLabel("Export all");
		allLogsToProM.setAlignmentX(Component.CENTER_ALIGNMENT);
		allNetsToProM.setAlignmentX(Component.CENTER_ALIGNMENT);
		exportLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		allLogsToProM.addActionListener(new AllLogsExportAction());
		allNetsToProM.addActionListener(new AllNetsExportAction());
		bottom_panel.add(exportLabel);
		bottom_panel.add(allLogsToProM);
		bottom_panel.add(allNetsToProM);
	}

	/**
	 * Initialise the top panel.
	 */
	private void initTopPanel() {
		top_panel.setLayout(new BoxLayout(top_panel, BoxLayout.Y_AXIS));
		top_panel.setOpaque(false);
		showClusterInfo(null);
	}

	/**
	 * Create a new JPanel in which the given key/value-pairs are visualised.
	 * Each entry in the list pairs will be mapped to one line.
	 * 
	 * Every Object[] in the list pairs should have length 2. If this is not the
	 * case, runtime exceptions may occur. The only two types of arrays accepted
	 * are String[] and Component[].
	 * 
	 * If an entry in pairs is of the type String[], then JLabels will be
	 * created for both strings in that array. Otherwise, the type Component[]
	 * is assumed.
	 * 
	 * If both the first and second part of an array-entry are not null, then
	 * the line corresponding to that entry will have the first part of the
	 * array (the key) on the left side and the second part of the array (the
	 * value) on the right side in a table-like layout. If the second part of
	 * the array is null, then the first part will get the entire line instead
	 * of a single cell.
	 * 
	 * @param pairs
	 *            The key-value pairs to visualise.
	 * @return The visualisation of the given pairs.
	 */
	public JPanel newKeyValuePanel(List<Object[]> pairs) {
		JPanel panel = new JPanel();
		panel.setOpaque(false);
		panel.setLayout(new GridBagLayout());
		GridBagConstraints c1 = new GridBagConstraints();
		GridBagConstraints c2 = new GridBagConstraints();
		c1.anchor = GridBagConstraints.LINE_START;
		c2.anchor = GridBagConstraints.LINE_START;
		c1.fill = GridBagConstraints.HORIZONTAL;
		c2.fill = GridBagConstraints.HORIZONTAL;
		c1.gridx = 0;
		c2.gridx = 1;
		c1.insets = new Insets(0, 0, 0, 10);
		c2.weightx = 1.0;
		int line = 0;
		for (Object[] pair : pairs) {
			c1.gridy = line;
			c2.gridy = line;
			if (pair instanceof String[]) {
				String[] pairstr = (String[]) pair;
				if (pair[0] != null) {
					if (pair[1] != null) {
						panel.add(newLabel(pairstr[0]), c1);
					} else {
						c1.gridwidth = 2;
						panel.add(newLabel(pairstr[0]), c1);
						c1.gridwidth = 1;
					}
				}
				if (pair[1] != null)
					panel.add(newLabel(pairstr[1]), c2);
			} else if (pair instanceof Component[]) {
				Component[] paircmp = (Component[]) pair;
				if (pair[0] != null) {
					if (pair[1] != null) {
						panel.add(paircmp[0], c1);
					} else {
						c1.gridwidth = 2;
						panel.add(paircmp[0], c1);
						c1.gridwidth = 1;
					}
				}
				if (pair[1] != null)
					panel.add(paircmp[1], c2);
			}
			line++;
		}
		return panel;
	}

	/**
	 * Create a new JLabel with the correct style to use in this side panel.
	 * 
	 * @param text
	 *            The text to display on the newly created label.
	 * @return A JLabel with correct layout and text.
	 */
	private JLabel newLabel(String text) {
		JLabel label = new JLabel(text);
		label.setForeground(foreground);
		label.setFont(label.getFont().deriveFont(Font.PLAIN));
		return label;
	}

	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		float deltaIntensity = 255f / (30f * 30f);
		for (int i = 0; i < 30; i++) {
			// Nearly as grey as Letherme.
			int alpha = (int) (deltaIntensity * i * i);
			g2.setColor(new Color(50, 50, 50, alpha));
			g2.fillRect(i, 0, 1, this.getHeight());
		}
		g2.setColor(new Color(50, 50, 50, 255));
		g2.fillRect(30, 0, this.getWidth() - 30, this.getHeight());
	}

	/**
	 * Change the selected JClusterBox to the given one.
	 */
	public void setSelectedBox(JClusterBox box) {
		if (selected != null)
			selected.setBorder(BorderFactory.createLineBorder(Color.BLACK, 2));
		if (box != null)
			box.setBorder(BorderFactory.createLineBorder(new Color(240, 100,
					100), 2));
		selected = box;
	}

	/**
	 * Show detailed information on the given ClusterStatistics.
	 */
	public void showClusterInfo(ClusterStatistics statistics) {
		top_panel.removeAll();
		if (statistics == null) {
			top_panel.add(newLabel("Click any cluster to get more"));
			top_panel.add(newLabel("information about it or to export"));
			top_panel.add(newLabel("any of the cluster's data."));
		} else {
			SlickerFactory slicker = SlickerFactory.instance();
			Font subtitle = this.getFont();
			subtitle = subtitle.deriveFont(Font.BOLD, 13f);

			JLabel eli = newLabel("Event Log Information");
			eli.setFont(subtitle);
			top_panel.add(eli);
			eli.setAlignmentX(Component.CENTER_ALIGNMENT);
			ArrayList<Object[]> eliInfo = new ArrayList<Object[]>();
			eliInfo.add(new String[] { "traces",
					Integer.toString(statistics.getNbTraces()) });
			eliInfo.add(new String[] { "dpi",
					Integer.toString(statistics.getNbDpi()) });
			eliInfo.add(new String[] { "events",
					Integer.toString(statistics.getNbEvents()) });
			eliInfo.add(new String[] { "mean trace length",
					Float.toString(statistics.getMeanTraceLength()) });
			eliInfo.add(new String[] { "activity classes",
					Integer.toString(statistics.getNbInternalDistinctEvents()) });
			eliInfo.add(new String[] { "extra conformant traces",
					Integer.toString(statistics.getNbExtraConformantTraces()) });
			eliInfo.add(new String[] {
					"extra non-conformant traces",
					Integer.toString(statistics.getNbExtraNonConformantTraces()) });
			top_panel.add(newKeyValuePanel(eliInfo));

			top_panel.add(Box.createVerticalStrut(20));

			JLabel pmi = newLabel("Process Model Information");
			pmi.setFont(subtitle);
			top_panel.add(pmi);
			pmi.setAlignmentX(Component.CENTER_ALIGNMENT);
			ArrayList<Object[]> pmiInfo = new ArrayList<Object[]>();
			JLabel pmiAccuracy = newLabel("Accuracy");
			pmiAccuracy.setFont(pmiAccuracy.getFont().deriveFont(Font.ITALIC));
			pmiInfo.add(new Component[] { pmiAccuracy, null });
			pmiInfo.add(new String[] { "ICS-fitness",
					Float.toString(statistics.getCluster().getFitness()) });
			pmiInfo.add(new String[] { "PM", Float.toString(statistics.getPM()) });
			pmiInfo.add(new String[] { "fitting dpi",
					Integer.toString(statistics.getNbFittingDpi()) });
			JLabel pmiComp = newLabel("Comprehensibility");
			pmiComp.setFont(pmiComp.getFont().deriveFont(Font.ITALIC));
			pmiInfo.add(new Component[] { pmiComp, null });
			pmiInfo.add(new String[] { "nodes",
					Integer.toString(statistics.getNbNodes()) });
			pmiInfo.add(new String[] { "arcs",
					Integer.toString(statistics.getNbArcs()) });
			pmiInfo.add(new String[] { "Average Connector Degree",
					Float.toString(statistics.getAverageConnectorDegree()) });
			pmiInfo.add(new String[] { "density",
					Float.toString(statistics.getNetDensity()) });
			JLabel pmiNodes = newLabel("Nodes");
			pmiNodes.setFont(pmiNodes.getFont().deriveFont(Font.ITALIC));
			pmiInfo.add(new Component[] { pmiNodes, null });
			{
				ArrayList<Object[]> nodes = new ArrayList<Object[]>();
				List<XEvent> events = statistics.getInternalDistinctEvents();
				String transition = null;
				boolean allthesame = true;
				for (XEvent event : events) {
					XAttributeMap attributes = event.getAttributes();
					nodes.add(new String[] {
							attributes.get("concept:name").toString(),
							attributes.get("lifecycle:transition").toString() });
					if (transition != null && allthesame) {
						if (!attributes.get("lifecycle:transition").toString()
								.equals(transition))
							allthesame = false;
					} else {
						transition = attributes.get("lifecycle:transition")
								.toString();
					}
				}
				Collections.sort(nodes, new Comparator<Object[]>() {

					@Override
					public int compare(Object[] one, Object[] two) {
						String allone = one[0] + "\0" + one[1];
						String alltwo = two[0] + "\0" + two[1];
						return allone.compareTo(alltwo);
					}
				});
				if (allthesame) {
					for (Object[] node : nodes) {
						node[1] = null;
					}
				}
				JPanel nodePanel = newKeyValuePanel(nodes);
				JScrollPane scrollpane = new JScrollPane();
				JViewport viewport = new JViewport();
				scrollpane.setViewport(viewport);
				viewport.add(nodePanel);
				viewport.setOpaque(false);
				scrollpane.setOpaque(false);
				scrollpane.setPreferredSize(new Dimension(200, 100));
				pmiInfo.add(new Component[] { scrollpane, null });
			}
			top_panel.add(newKeyValuePanel(pmiInfo));

			top_panel.add(Box.createVerticalStrut(20));

			JLabel exportLabel = newLabel("Export data");
			exportLabel.setFont(subtitle);
			JButton buttonExportLog = slicker.createButton("Log to ProM");
			JButton buttonExportNet = slicker
					.createButton("HeuristicsNet to ProM");
			exportLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
			buttonExportLog.setAlignmentX(Component.CENTER_ALIGNMENT);
			buttonExportNet.setAlignmentX(Component.CENTER_ALIGNMENT);
			top_panel.add(exportLabel);
			top_panel.add(buttonExportLog);
			top_panel.add(buttonExportNet);

			top_panel.add(Box.createVerticalStrut(20));

			JButton buttonViewNet = slicker.createButton("View HeuristicsNet");
			buttonViewNet.setAlignmentX(Component.CENTER_ALIGNMENT);
			top_panel.add(buttonViewNet);

			buttonExportLog.addActionListener(new ExportLogAction(statistics
					.getCluster()));
			buttonExportNet.addActionListener(new ExportNetAction(statistics
					.getCluster()));
			buttonViewNet.addActionListener(new ViewNetAction(statistics
					.getCluster()));
		}
		top_panel.validate();
		this.validate();
	}
}
