package org.processmining.plugins.actitrac.ui.clustertree;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.processmining.plugins.actitrac.algorithm.ClusterStatistics;
import org.processmining.plugins.actitrac.types.Cluster;
import org.processmining.plugins.actitrac.types.GroupedXLog;
import org.processmining.plugins.actitrac.ui.JClusteringSidePane;


/**
 * A graphical component that represents a single Cluster. This component will
 * display some basic information on this cluster, occasionally in comparison to
 * the original log. When this component is clicked, it will also notify a given
 * JClusteringSidePane of this. This JClusteringSidePane might then, for example, show
 * more detailed information on this cluster.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-08-01
 */
public class JClusterBox extends JPanel {

	private static final long serialVersionUID = 5530873294008267921L;

	/**
	 * A ClusterStatistics object representing the Cluster this JClusterBox is
	 * representing.
	 */
	/* repreception */
	private ClusterStatistics statistics;

	/**
	 * The JClusteringSidePane to which we should report.
	 */
	private JClusteringSidePane infopane;

	/**
	 * This will notify the infopane when this component is clicked.
	 */
	private MouseListener jClusterBoxMouseListener = new MouseListener() {

		@Override
		public void mouseClicked(MouseEvent e) {
		}

		@Override
		public void mouseEntered(MouseEvent arg0) {

		}

		@Override
		public void mouseExited(MouseEvent arg0) {

		}

		@Override
		public void mousePressed(MouseEvent arg0) {

		}

		@Override
		public void mouseReleased(MouseEvent e) {
			// If still in this component
			if (e.getX() < 0 || e.getX() >= JClusterBox.this.getWidth())
				return;
			if (e.getY() < 0 || e.getY() >= JClusterBox.this.getHeight())
				return;
			infopane.setSelectedBox(JClusterBox.this);
			infopane.showClusterInfo(statistics);
			e.consume();
		}
	};

	/**
	 * Create a new JClusterBox for the given Cluster. The given GroupedXLog
	 * should be the original log, of which the given Cluster is part. The given
	 * JClusteringSidePane will be notified when this JClusterBox is clicked, so
	 * that it may take appropriate action.
	 * 
	 * @param cluster
	 *            The Cluster to represent.
	 * @param parentlog
	 *            The log of which this Cluster is a part.
	 * @param infopane
	 *            The JClusteringSidePane which must be notified when this
	 *            JClusterBox is clicked.
	 */
	public JClusterBox(Cluster cluster, GroupedXLog parentlog,
			JClusteringSidePane infopane) {
		this.infopane = infopane;
		this.statistics = new ClusterStatistics(cluster);
		this.setBorder(BorderFactory.createLineBorder(Color.BLACK, 2));
		JPanel contents = new JPanel();
		contents.setOpaque(false);
		this.add(contents);
		contents.setBorder(BorderFactory.createEmptyBorder(5, 10, 5, 10));
		contents.setLayout(new GridBagLayout());

		// name
		GridBagConstraints c = new GridBagConstraints();
		c.weightx = 1.0;
		c.gridwidth = 2;
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.BASELINE_LEADING;
		c.insets = new Insets(0, 0, 10, 0);
		String nameStr = cluster.getName();
		if (nameStr.lastIndexOf('_') > 20) {
			String sub1 = nameStr.substring(0, 16);
			String sub2 = nameStr.substring(nameStr.lastIndexOf('_'));
			nameStr = sub1 + ".." + sub2;
		}
		JLabel name = new JLabel(nameStr);
		Font title = name.getFont().deriveFont(Font.BOLD | Font.ITALIC, 13.0f);
		name.setFont(title);
		contents.add(name, c);

		// whitespace

		GridBagConstraints c1 = new GridBagConstraints();
		c1.anchor = GridBagConstraints.BASELINE_LEADING;
		c1.gridx = 0;
		c1.ipadx = 10;

		GridBagConstraints c2 = new GridBagConstraints();
		c2.anchor = GridBagConstraints.BASELINE_LEADING;
		c2.gridx = 1;

		// traces #
		int tracesNb = statistics.getNbTraces();
		float tracesPr = (tracesNb * 10000) / parentlog.getNbTraces() / 100f;
		c1.gridy = 1;
		c2.gridy = 1;
		contents.add(new JLabel("traces"), c1);
		contents.add(new JLabel(tracesNb + " (" + tracesPr + "%)"), c2);

		c1.insets = new Insets(0, 0, 5, 0);
		c2.insets = c1.insets;

		// dpi #
		int dpiNb = statistics.getNbDpi();
		float dpiPr = (dpiNb * 10000) / parentlog.size() / 100f;
		c1.gridy = 2;
		c2.gridy = 2;
		contents.add(new JLabel("dpi"), c1);
		contents.add(new JLabel(dpiNb + " (" + dpiPr + "%)"), c2);

		// whitespace
		c1.insets = new Insets(0, 0, 0, 0);
		c2.insets = c1.insets;

		// ICS-fitness #
		c1.gridy = 3;
		c2.gridy = 3;
		contents.add(new JLabel("ICS-fitness"), c1);
		contents.add(new JLabel(Float.toString(cluster.getFitness())), c2);

		// PM #
		c1.gridy = 4;
		c2.gridy = 4;
		contents.add(new JLabel("PM"), c1);
		contents.add(new JLabel(Float.toString(statistics.getPM())), c2);

		// fitting dpi #
		c1.gridy = 5;
		c2.gridy = 5;
		contents.add(new JLabel("fitting dpi"), c1);
		contents.add(
				new JLabel(Integer.toString(statistics.getNbFittingDpi())), c2);

		// // CS-fitness #
		// ContinuousSemantics cs = new
		// ContinuousSemantics(XLogInfoFactory.createLogInfo(parentlog.getLog()));
		// HeuristicsNet net =
		// cluster.getProcessModel().getHeuristicsNet().copy();
		// net = cs.calculate(new HeuristicsNet[] {net})[0];
		// c1.gridy = 6;
		// c2.gridy = 6;
		// contents.add(new JLabel("CS-fitness"), c1);
		// contents.add(new JLabel(Float.toString((float) net.getFitness())),
		// c2);

		int alt_color = (int) (200 * (1f - tracesPr / 100f)) + 40;
		this.setBackground(new Color(alt_color, 240, alt_color));
		for (Component comp : this.getComponents()) {
			comp.setForeground(Color.BLACK);
		}
		name.setForeground(Color.WHITE);

		this.addMouseListener(jClusterBoxMouseListener);
	}

	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;

		// Calculate the difference between the background and the target colour
		// (some very light gray). Then divide this difference through the
		// amount of units over which we want to spread the change (the width of
		// the component, in pixels).
		// Eg. every 1 pixel, this deltaPix should be added to the colour.
		Color back = this.getBackground();
		int[] colorLeft = new int[] { back.getRed(), back.getGreen(),
				back.getBlue() };
		int[] colorRight = new int[] { 240, 240, 240 };
		float[] deltaPix = new float[3];
		for (int i = 0; i < 3; i++) {
			deltaPix[i] = (colorRight[i] - colorLeft[i])
					/ (float) this.getWidth();
		}

		// Pretty much the same for the title, only this time we just want to
		// darken it quite a bit. And we want a cubic scale, so we divide
		// through the amount of pixels over which we want to spread the change
		// squared.
		int[] colorTitle = new int[] { -150, -150, -150 };
		float[] titleDeltaSq = new float[3];
		for (int i = 0; i < 3; i++) {
			titleDeltaSq[i] = colorTitle[i] / (float) (50 * 50);
		}
		
		int[] currentColor = new int[3];
		for (int i = 0; i < this.getWidth(); i++) {
			// Calculate the current colour for this (vertical) line of pixels.
			for (int j = 0; j < 3; j++)
				currentColor[j] = (int) (colorLeft[j] + deltaPix[j] * i);
			g2.setColor(new Color(currentColor[0], currentColor[1],
					currentColor[2]));
			// Draw the vertical line of pixels.
			g2.drawRect(i, 0, 1, this.getHeight());
			
			// Draw the fancy title bar
			// Part 1: light to dark (slow)
			int[] basecolor = new int[] { currentColor[0], currentColor[1],
					currentColor[2] };
			for (int j = 0; j < 30; j++) {
				for (int k = 0; k < 3; k++) {
					int p = j + 20;
					p *= p;
					currentColor[k] = (int) (colorLeft[k] + p * titleDeltaSq[k]);
					if (currentColor[k] < 0)
						currentColor[k] = 0;
				}
				g2.setColor(new Color(currentColor[0], currentColor[1],
						currentColor[2]));
				g2.drawRect(i, j, 1, 1);
			}
			// Part 2: dark to light (fast)
			for (int j = 0; j < 5; j++) {
				for (int k = 0; k < 3; k++) {
					int p = 50 - 10 * j;
					p *= p;
					currentColor[k] = (int) (basecolor[k] + p * titleDeltaSq[k]);
					if (currentColor[k] < 0)
						currentColor[k] = 0;
				}
				g2.setColor(new Color(currentColor[0], currentColor[1],
						currentColor[2]));
				g2.drawRect(i, 30 + j, 1, 1);
			}
		}
	}
}
