package org.processmining.plugins.actitrac.types;

import java.util.Iterator;

/**
 * This class represents a tree of clusters. The actual structure of the data is
 * more a linked list, which will also be exploited by this class, but for the
 * sake of generalisation is it represented as a tree. This type of tree
 * consists of three types of nodes: ClusterTreeLeaf-nodes,
 * ClusterTreeInnerNode-nodes and one ClusterTreeRoot.
 * 
 * This tree should always consist of at least two leaves. The behaviour of this
 * class is undefined when this requirement is not met.
 * 
 * The tree leans to the left. The left child of each node (apart of the leaves
 * and the last inner node) should be an inner node. The right child of each
 * node (not the leaves) should be a leaf. Finally, the left child of the last
 * inner node is also a leaf.
 * 
 * This implementation also allows iterating over the Clusters in the tree. You
 * can thus use an enhanced for loop to iterate over each cluster in the tree.
 * This iteration will not include the original (root) cluster.
 * 
 * @author Niels Lambrigts (niels.lambrigts@gmail.com)
 * @author Jochen De Weerdt (Jochen.DeWeerdt@econ.kuleuven.be)
 * @date 2011-07-29
 */
public class ClusterTree implements Iterable<Cluster> {

	/**
	 * The root of the tree.
	 */
	private ClusterTreeRoot root;

	/**
	 * Create a new ClusterTree. The given cluster should represent the original
	 * (non-clustered) input log.
	 * 
	 * @param pseudoCluster
	 *            A Cluster object representing the original input.
	 */
	public ClusterTree(Cluster pseudoCluster) {
		root = new ClusterTreeRoot(pseudoCluster);
	}

	/**
	 * Return the root of the tree.
	 * 
	 * @see ClusterTree for more information on the way a ClusterTree is
	 *      structured.
	 */
	public ClusterTreeRoot getRoot() {
		return root;
	}

	/**
	 * Add a Cluster to the bottom of this tree.
	 * 
	 * @see ClusterTree for more information on the way a ClusterTree is
	 *      structured.
	 */
	public void addCluster(Cluster cluster) {
		ClusterTreeNode add = new ClusterTreeLeaf(cluster);
		if (root.getRight() == null) {
			root.setRight(add);
			return;
		}
		if (root.getLeft() == null) {
			root.setLeft(add);
			return;
		}
		ClusterTreeNode current = root;
		while (!(current.getLeft() instanceof ClusterTreeLeaf))
			current = current.getLeft();
		ClusterTreeInnerNode addInner = new ClusterTreeInnerNode();
		addInner.setRight(current.getLeft());
		addInner.setLeft(add);
		((ClusterTreeInnerNode) current).setLeft(addInner);
	}

	@Override
	public Iterator<Cluster> iterator() {
		Iterator<Cluster> iterator = new Iterator<Cluster>() {

			ClusterTreeNode current = root;

			@Override
			public boolean hasNext() {
				return current != null;
			}

			@Override
			public Cluster next() {
				if (!(current instanceof ClusterTreeLeaf)) {
					Cluster out = ((ClusterTreeLeaf) current.getRight())
							.getCluster();
					current = current.getLeft();
					return out;
				} else {
					Cluster out = ((ClusterTreeLeaf) current).getCluster();
					current = null;
					return out;
				}
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
		return iterator;
	}
}
